import boto3
import time
import pandas as pd
import requests
import io

from requests.auth import HTTPBasicAuth
import logging

logger = logging.getLogger(__name__)


class Boto3ClientException(Exception):
    """Exception raised for errors in the input to Boto3ClientFactory methods.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, expression, message):
        self.expression = expression
        self.message = message

class Boto3ClientFactory:
    """
    Create boto S3 client, possibly using Lilly's API gateway that provides temporary credentials
    Full documentation to be written
    """
    __service_account = ""
    __aws_account_id = ""
    __gw_s3_access_role = ""
    __cert_path = ""
    __cert_key_path = ""
    __gw_url = ""
    __proxy_dict = {}
    __role_arn = ""
    
    def set_service_account(self, service_account):
        self.__service_account = service_account
        return self
    
    def set_aws_account_id(self, aws_account_id):
         self.__aws_account_id = aws_account_id
         return self
        
    def set_aws_access_role(self, gw_s3_access_role):
        self.__gw_s3_access_role = gw_s3_access_role
        return self
        
    def set_certificate_path(self, cert_path):
        self.__cert_path = cert_path
        return self
    
    def set_certificate_key_path(self, cert_key_path):
        self.__cert_key_path = cert_key_path
        return self
    
    def set_gw_url(self, gw_url):
        self.__gw_url = gw_url
        return self
        
    def set_proxies(self, http_proxy, https_proxy, ftp_proxy):
        self.__proxy_dict = { 
              "http"  : http_proxy, 
              "https" : https_proxy, 
              "ftp"   : ftp_proxy
            }
        return self
    
    def _validate(self):
        if self.__service_account == "":
            raise Boto3ClientException("service_account", "PLEASE PROVIDE A VALID AWS SERVICE ACCOUNT")
        
        if self.__aws_account_id == "":
            raise Boto3ClientException("aws_account_id", "PLEASE PROVIDE A VALID AWS ACCOUNT ID")
        
        if self.__gw_s3_access_role == "":
            raise Boto3ClientException("gw_s3_access_role", "PLEASE PROVIDE A VALID AWS GATEWAY ACCESS ROLE")
            
        if self.__cert_path == "":
            raise Boto3ClientException("cert_path", "PLEASE PROVIDE A VALID CERTIFICATE PATH")
        
        if self.__cert_key_path == "":
            raise Boto3ClientException("cert_key_path", "PLEASE PROVIDE A VALID CERTIFICATE KEY PATH")
        
        if self.__gw_url == "":
            raise Boto3ClientException("gw_url", "PLEASE PROVIDE A VALID GATEWAY URL")
        
    def create_using_gw(self, service_account_pwd, client_type='s3', **kwargs):
        self._validate()
        self.__role_arn = 'arn:aws:sts::' + self.__aws_account_id + ':assumed-role/' + self.__gw_s3_access_role + '/' + self.__service_account
        
        authorization=HTTPBasicAuth(self.__service_account,service_account_pwd)
        cert = (self.__cert_path, self.__cert_key_path)
        headers = {'USER': 'application/json','content-type': 'application/json'}
        try:
            temporary_credentials_response=requests.get(self.__gw_url, headers=headers, auth=authorization, cert=cert, proxies=self.__proxy_dict)
            for line in temporary_credentials_response.iter_lines():
                logger.debug('>>>Temp credentials: {}'.format(line))
                if line:
                    decoded_line = line.decode('utf-8')
                    resjson = temporary_credentials_response.json()
                    temporary_credentials = temporary_credentials_response.json()
                    credentials = temporary_credentials
            for key, value in credentials.items():
                try:
                    assume_role_with_saml_result = credentials[key]['AssumeRoleWithSAMLResponse']['AssumeRoleWithSAMLResult']
                    if assume_role_with_saml_result['AssumedRoleUser']['Arn']==self.__role_arn:
                        credentials_field = assume_role_with_saml_result['Credentials']
                        session_token = credentials_field['SessionToken']
                        aws_access_key = credentials_field['AccessKeyId']
                        aws_secret_access_key = credentials_field['SecretAccessKey']
                except KeyError:
                    pass  # we only want assumerolewithsamlresponse
            credentials_field = assume_role_with_saml_result['Credentials']
            session_token = credentials_field['SessionToken']
            aws_access_key = credentials_field['AccessKeyId']
            aws_secret_access_key = credentials_field['SecretAccessKey']
            logger.debug("SessionToken {}\nAccessKeyId {}\nSecretAccessKey {}".format(session_token, aws_access_key, aws_secret_access_key))
            s3 = boto3.client(client_type, aws_access_key_id=aws_access_key, aws_secret_access_key=aws_secret_access_key, aws_session_token=session_token, **kwargs)
        except KeyError:
            logger.error("Error Found: " + KeyError)
        return s3

    def create(self, aws_id = "", aws_secret = "", client_type='s3', **kwargs):
        if len(aws_id) > 0:
            s3 = boto3.client(client_type, aws_access_key_id=aws_id, aws_secret_access_key=aws_secret, **kwargs)
        else:
            s3 = boto3.client(client_type, **kwargs)
        return s3

    
class S3SelectHelper():
    def __init__(self, s3_boto_client):
        self.__s3 = s3_boto_client
        self.set_quote_character()
        self.set_separator()
    
    def set_quote_character(self, quote_character = "'"):
        self.__quote_character = quote_character
    
    def set_separator(self, separator = ","):
        self.__separator = separator
        
    def s3_select_to_df(self, bucket, key, expression, headers):
        logger.info('Quote delimiter set to: [{}]'.format(self.__quote_character))
        start = time.time()
        r, cnt = self.__execute_s3_select(bucket, key, expression)
        end = time.time()
        logger.info('Read {0} rows in {1:.2f} seconds'.format(str(cnt), end - start))
        r = headers + "\n" + r
        return pd.read_csv(io.StringIO(r), sep=self.__separator)
    
    def __execute_s3_select(self, bucket, key, expression):
        compression_type = S3SelectHelper.infer_compression_by_filename(key)
        logger.info("Compression: {}".format(compression_type))
        r = self.__s3.select_object_content(
                Bucket=bucket,
                Key=key,
                ExpressionType='SQL',
                Expression=expression,
                InputSerialization = {'CSV': {"FileHeaderInfo": "Use", "AllowQuotedRecordDelimiter": True, "QuoteCharacter": self.__quote_character}, 'CompressionType': compression_type},
                OutputSerialization = {'CSV': {}},
        )
        event_stream = r['Payload']
        end_event_received = False
        records = ""
        logger.info("Appending rows...")
        cnt = 0
        for event in event_stream:
            cnt = cnt + 1
            if 'Records' in event:
                if cnt%1000 == 0:
                    logger.info("...read 1000 blocks...{}".format(cnt))

                records = records + event['Records']['Payload'].decode('utf-8')

            if 'End' in event:
                end_event_received = True

        if not end_event_received:
            raise Exception("End event not received, request incomplete.")

        return records.rstrip(), records.count("\n")
    
    @staticmethod
    def infer_compression_by_filename(filename):
        compression_type = 'NONE'
        if filename.endswith('.gz'):
            compression_type = 'GZIP'
        return compression_type


"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

__version__ = '1.1.2'


"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

import pylab as pl
import numpy as np
from tslearn.metrics import dtw_path


def tam(path):
    """
    Calculates the Time Alignment Measurement (TAM) based on an optimal warping path
    between two time series.
    Reference: Folgado et. al, Time Alignment Measurement for Time Series, 2018.

    :param path: (ndarray)
                A nested array containing the optimal warping path between the
                two sequences.
    :param report: (string)
                A string containing the report mode parameter.
    :return:    In case ``report=instants`` the number of indexes in advance, delay and phase
                will be returned. For ``report=ratios``, the ratio of advance, delay and phase
                will be returned. In case ``report=distance``, only the TAM will be returned.

    """
    # Delay and advance counting
    delay = np.sum(np.diff(path[0]) == 0)
    advance = np.sum(np.diff(path[1]) == 0)

    # Phase counting
    incumbent = np.sum(np.diff(path[0]) == 1) * (np.diff(path[1]) == 1)
    phase = len(incumbent)

    # Estimated and reference time series duration.
    len_estimation = path[1][-1]
    len_ref = path[0][-1]

    p_advance = advance * 1. / len_ref
    p_delay = delay * 1. / len_estimation
    p_phase = phase * 1. / np.min([len_ref, len_estimation])

    return p_advance + p_delay + (1 - p_phase)


def tam_distance(seriesA, seriesB):
    
    d = dtw_path(seriesA,seriesB)[0]
    p = np.asarray([[i[0] for i in d],[i[1] for i in d]])
    return tam(p)


"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

"""
Sub-module of TSAnalysis

Contains classes and methods to perform preprocessing on signal data
"""

import numpy as np
import pandas as pd
import logging
import random
from sklearn.preprocessing import MinMaxScaler
from tqdm import tqdm
from deprecation import deprecated

logger = logging.getLogger(__name__)


class DataPreparation:
    def __init__(self):
        self.__time_fld = 'vtime'
        self.__name_fld = 'vtag'
        self.__value_fld = 'vvalue'


    def setup_mappings(self, time_fld, name_fld, value_fld):
        self.__time_fld = time_fld
        self.__name_fld = name_fld
        self.__value_fld = value_fld


    def get_feature_matrix(self, df, feature_list):
        try:
            feat_df = df.pivot_table(index=self.__time_fld, columns=self.__name_fld, values=self.__value_fld)
            feat_df_cols = list(feat_df.columns)
            diff_cols = self.__get_diff(feature_list, feat_df_cols)
            missing_cols_df = pd.DataFrame(index=feat_df.index, columns=diff_cols)
            feat_df = pd.concat([feat_df, missing_cols_df], axis=1)
            feat_df = feat_df[feature_list]
            return feat_df
        except Exception as error:
            logger.info("Unable to create feature matrix :{}".format(error))
    
    # Fixing SHAP error, will be removed once entire project moves to new schema
    def get_feature_matrix_Old(self, df, feature_list):
        try:
            feat_df = df.pivot_table(index='event_time', columns='ft_name', values='nvalue')
            feat_df_cols = list(feat_df.columns)
            diff_cols = self.__get_diff(feature_list, feat_df_cols)
            missing_cols_df = pd.DataFrame(index=feat_df.index, columns=diff_cols)
            feat_df = pd.concat([feat_df, missing_cols_df], axis=1)
            feat_df = feat_df[feature_list]
            return feat_df
        except Exception as error:
            logger.info("Unable to create feature matrix :{}".format(error))
    
    def check_data_quality(self, df, feature_list, max_null_ratio):
        quality = False
        try:
            cycles = len(df[self.__time_fld].unique())
            existing_features = df[self.__name_fld].unique()
            feature_match1 = set(df[self.__name_fld].unique()).issubset(feature_list)
            feature_match2 = set(feature_list).issubset(df[self.__name_fld].unique())
            df_feat = self.get_feature_matrix(df, feature_list)
            attributr_nan_percentage = self.__get_naratio(df_feat, feature_list)
            null_flag = max(attributr_nan_percentage) < max_null_ratio
            null_validation = self.__validate(df_feat, feature_list, thr=max_null_ratio)

            if feature_match1 and feature_match2 and null_flag and null_validation:
                quality = True
            
            data_quality_report = {}
            data_quality_report['cycle count'] = cycles
            data_quality_report['data set contains all features'] = feature_match1 and feature_match2 
            data_quality_report['data set contains all features'] = feature_match2 
            data_quality_report['feature list'] = existing_features
            data_quality_report['null feature percentage'] = attributr_nan_percentage
            data_quality_report['quality pass'] = quality
            return [quality, data_quality_report]
        except Exception as error:
            logger.info("Unable to generate quality :{}".format(error))
            
    
    def normalize(self,df):
        try:
            scaler = MinMaxScaler()
            scaler.fit(df)
            return [scaler, scaler.transform(df)]
        except Exception as error:
            logger.info("Unable to Normalize data :{}".format(error))

            
    def fill_na(self,df):
        try:
            df_new = df.apply(lambda x: x.fillna(x.mean()), axis=0)
            return df_new
        except Exception as error:
            logger.info("Error in imputation :{}".format(error))

    
    def __get_diff(self, li1, li2):                                            # make private 
        return (list(list(set(li1)-set(li2)) + list(set(li2)-set(li1))))

    
    def __get_naratio(self, df_feat, feature_list):
        try:
            nacounts = []
            for ftn in feature_list:
                nacounts.append(df_feat[ftn].isna().sum() / len(df_feat))
            return nacounts
        except Exception as error:
            logger.info("Error while calculating null values :{}".format(error))

            
    def __validate(self, df_feat, feature_list, thr = 0.4):
        try:
            nacounts = self.__get_naratio(df_feat, feature_list)
            for v in nacounts:
                if (v > thr):
                    return False
            return True
        except Exception as error:
            logger.info("Unable to validate :{}".format(error))

    
    @staticmethod
    def measure_distances_cycles(s1, s2, distances_dict):
        distance_values = []
        for dist_name in distances_dict:
            f = distances_dict[dist_name]
            distance_values.append(f.compute_distance_(s1, s2))
        return distance_values
        
            
    @staticmethod
    @deprecated(deprecated_in="0.8.0", removed_in="0.9.0", details="Use CyclePerturbationHelper.perturb_cycles(...) from perturbations.py instead")
    def perturb_cycles(df, cycle_col, tags_col, signal_col, perturbations_dict, tags_list = 'all'):
        df_test = df.copy()
        perturb_list = []
        cycles_list = df_test[cycle_col].unique()

        if tags_list == 'all':
            tags_list_new = df_test[tags_col].unique()
        else:
            tags_list_new = list(set(df_test[tags_col].unique()) & set(tags_list))
            
        for cycle in cycles_list:
            df_cycle = df_test[df_test[cycle_col] == cycle]
            for tag in tags_list_new:
                idxvals = df_cycle[df_cycle[tags_col] == tag].index
                tmp_arr = df_cycle.loc[idxvals, (signal_col)].values
                for pert_name in perturbations_dict:
                    pert_class = perturbations_dict[pert_name]
                    tmp_arr = pert_class.perturb_transform_(tmp_arr)            
                df_test.loc[idxvals, (signal_col)] = tmp_arr
        return df_test
    
    
    @staticmethod
    def create_tag_distance_columns(tags_list, distances_list):
        tag_distance_cols = []
        for tag in tags_list:
            for distance in distances_list:
                new_col = tag + "_" + distance
                tag_distance_cols.append(new_col)
                
        return tag_distance_cols
    
    @staticmethod
    def create_distance_matrix(df, golden_cycle_df, golden_ref_no = None, 
                               cycle_col = None, tags_col = None, signal_col = None, 
                               distances_dict = None, tags_list = None):
    
        cycles = set(df[cycle_col])
        master_table = []
        distances_list = list(distances_dict.keys())
        tag_distance_cols = DataPreparation.create_tag_distance_columns(tags_list = tags_list, distances_list = distances_list)
        for cycle in tqdm(cycles):
            if cycle == golden_ref_no:
                continue
                
            cycle_vals = []
            df_temp = df[df[cycle_col] == cycle]
            cycle_tags = df_temp[tags_col].unique()
            for tag in tags_list:
                if tag in cycle_tags:
                    ts1 = df_temp[df_temp[tags_col] == tag][signal_col].values
                    ts2 = golden_cycle_df[golden_cycle_df[tags_col] == tag][signal_col].values 
                    distances = DataPreparation.measure_distances_cycles(ts1, ts2, distances_dict=distances_dict)
                    cycle_vals.extend(distances)
                else:
                    cycle_vals.extend(repeat(np.nan, len(distances_dict.keys())))
            cycle_vals.append(cycle)
            master_table.append(cycle_vals)
	
        tag_distance_cols.append('cycle')
        master_df = pd.DataFrame(master_table, columns = tag_distance_cols)
        return master_df


# Defines the main classes
class SignalSegment:
    """
    Segment signal into sub-signals based on rules specified
    in a segmentation function (segmenting_rule_fn())
    """

    def __repr__(self):
        return """Segmentation of signal into sub-signals based on 
        rules specified in segmenting_rule_fn()
        """

    def get_segments(self, df, segmenting_rule_fn, signal_col, order=True, orderby_cols=['time', 'tag']):
        """
        Apply segmentation logics.
        
        Parameters
        ----------
        df: Pandas Data Frame
            Contains data structured using the following columns: tag, time, *signal_col*

        segmenting_rule_fn: function
            Function according to which splitting is done

        signal_col : str
            The "value" column contains time signal values that can 
            be split into sub-signal (cycles)

        order: boolean, optional (default is True)
            For the segmentation logics to work, data in df should be 
            ordered by "time". The method orders the data by default
            
        Returns
        ----------
        df: Pandas Data Frame
            Contains data as in the original data frame, with an added column 
            named "cycle" reporting the sub-signal number
        
        """
        
        try:
            df.drop(["svalue", "status", "flags"], axis = 1, inplace=True)
        except:
            pass
        
        df[signal_col] = pd.to_numeric(df[signal_col], errors='coerce')
        df.dropna(inplace=True)
        df.time = pd.to_datetime(df.time)
        
        if order:
            df.sort_values(by=orderby_cols, inplace=True)
        
        df = df.reset_index(drop=True)
        
        cycle_index = df[segmenting_rule_fn(df)].index
        logging.debug("Found {} breaking points. First three are:".format(len(cycle_index)))
        logging.debug("{}".format(df[df.index == cycle_index[0]].time))
        logging.debug("{}".format(df[df.index == cycle_index[1]].time))
        logging.debug("{}".format(df[df.index == cycle_index[2]].time))
        
        logging.debug(cycle_index)
        
        df['cycle'] = np.nan

        logging.debug(np.arange(1,len(cycle_index)+1))
        
        df.loc[cycle_index, ('cycle')] = np.arange(1,len(cycle_index)+1)
        df = df.fillna(method='ffill')
        df = df.fillna(0)
        df['cycle'] = df.cycle.astype('int')
        return df
        

class SignalSegmentClean:
    """
    Eliminate those signals with durations that are too short or too long
    """
    
    def __init__(self, tag_name_list):
        """
        Class constructor.
        
        Parameters
        ----------
        tag_name_list: string
            list of names of tags that class focuses on
        """
        self.__tag_name_list = tag_name_list


    def __repr__(self):
        return """ Cleaning signal with tag names '{}'
        """.format(self.__tag_name_list)
    
    
    def prune_segments(self, df, low_perc = 0.25, hi_perc = 0.75):
        """
        Apply cleaning logics. Discard sub-signals (cycles) that are too short or too long
        
        Parameters
        ----------
        df: Pandas Data Frame
            contains data structured using the following columns: 
            tag, time, value, cycle
            Any df passed through preprocessing.segmentation will follow 
            this canonical form

        low_perc: float, optional (default: 0.25)
            low percentile: cycles having duration lower than low_perc 
            are discarded from the data set
            
        hi_perc: float, optional (default: 0.75)
            high percentile: cycles having duration higher than hi_perc 
            are discarded from the data set
            
        Returns
        ----------
        df: Pandas Data Frame, containing only values related to 
            tag_name_list and cleaned up from cycles that are too short or too long
            
        """
        df_clean = df[df.tag.isin(self.__tag_name_list)]

        logger.info("""Aggregating all sub-signals and pruning at cutoff points. 
        This might take some time.""")
            
        control = df_clean.groupby(['cycle']).agg({ 'time' : [min, max]}).time
        control['tot'] = (control['max'] - control['min']).dt.total_seconds()
        logger.debug(control)
        logger.info("Cutting cycles having durations out of [{}, {}]".format(np.quantile(control.tot,low_perc), np.quantile(control.tot,hi_perc)))
        control_index = control[(control.tot<=np.quantile(control.tot,hi_perc)) 
                                & (control.tot>=np.quantile(control.tot,low_perc))].index.to_list()

        df_clean = df_clean[df_clean.cycle.isin(control_index)]
        return df_clean


# Defines the user interface
def segmentation(df, segmenting_rule_fn, signal_col='value', order = True, orderby_cols=['time', 'tag']):

    """
    Function that returns a DataFrame of signals segmented 
    from a longer signal using some specified rule(s)

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing the signal to be segmented

    segmenting_rule_fn: function
        Function according to which the segmentation is done
    
    signal_col: str, optional (default = 'value')
        The "value" column contains time signal values that can 
        be split into sub-signal (cycles)
    
    order : boolean, optional (default = True)
        For the segmentation logics to work, data in df should be 
        ordered by "time". The method orders the data by default

    Returns
    --------
    df: pandas.DataFrame
        Contains data as in the original data frame, with an added column 
        named "cycle" reporting the sub-signal number

    """

    segment_instance = SignalSegment()
    segmented_df = segment_instance.get_segments(df, segmenting_rule_fn,
                                                 signal_col, order=True, orderby_cols=orderby_cols)

    return segmented_df


def prune(df, tag_name_list, low_perc = 0.25, hi_perc = 0.75):
    """
    Function that prunes a signal segment based on the cutoff points

    Parameters
    -----------
    df: pandas.DataFrame
        Contains data structured using the following columns: 
        tag, time, value, cycle
        Any df passed through preprocessing.segmentation will follow 
        this canonical form

    tag_name_list: str
        List of names of tags pruning focuses on
    
    low_perc: float, optional (default: 0.25)
        Low percentile: cycles having duration lower than low_perc 
        are discarded from the data set
            
        hi_perc: float, optional (default: 0.75)
        High percentile: cycles having duration higher than hi_perc 
        are discarded from the data set
            
    Returns
    ----------
    df: pandas.DataFrame
        Containing only values related to tag_name_list and 
        cleaned up from cycles that are too short or too long

    """

    prune_instance = SignalSegmentClean(tag_name_list)
    pruned_df = prune_instance.prune_segments(df, low_perc, hi_perc)

    return pruned_df


def ts_train_test_split(df_ts, **kwargs):

    """
    Function that splits a dataframe into train and test components

    Parameters:
    -----------
        df_ts: pandas.DataFrame
            The input dataframe

        kwargs: dict
        Other arguments
            train_size: float, default = 0.8
                Fraction of the input dataframe to be allocated to the train split

            test_size: float, default = 0.2
                Fraction of the input dataframe to be allocated to the test split

            shuffle: bool, default = True
                Whether or not the input dataframe is shuffled

    Returns:
    ---------
        df_train: pandas.DataFrame
            The train split

        df_test: pandas.DataFrame
            The test split

    """
    
    train_size = kwargs.pop('train_size', None)
    test_size = kwargs.pop('test_size', None)
    shuffle = kwargs.pop('shuffle', True)
    
    default_train_size = 0.8
    
    if train_size is None and test_size is None:
        train_size = default_train_size
        
    if train_size and test_size:
        raise TypeError("Only one of train_size or test_size needs to be specified, not both")
    
    if kwargs:
        raise TypeError("Invalid parameters passed")
    
    cycles = df_ts.cycle.unique()
    n_cycles = len(cycles)
    
    if shuffle:
        random.shuffle(cycles)
        
    cutpoint = round(train_size * n_cycles)  
    
    df_train = df_ts[df_ts.cycle.isin(cycles[:cutpoint])]
    df_test = df_ts[df_ts.cycle.isin(cycles[cutpoint:])]
    
    return df_train, df_test


"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

import multiprocessing as mp
from aimltools.ts.distances import *
import itertools
import tslearn.barycenters


class GoldenCycle():
    
    def __init__(self):
        pass
    
    def get_closest_real_cycle(self, gc, df_tag, distance_metric):
        cycle_list = df_tag.cycle.unique()
        d_min = float('inf')
        grc_min = []
        cycle_min = -1
        logger.info("candidate cycles: {}".format(len(cycle_list)))
        for cycle_n in cycle_list:
            ts = df_tag[df_tag.cycle == cycle_n].value.values
            d = distance_metric.compute_distance_(gc, ts)
            if d < d_min:
                d_min = d
                grc_min = ts
                cycle_min = cycle_n

        logger.info("closest real cycle: {}".format(cycle_min))
        return grc_min

    def get_barycenter(self, df_tag, barycenter_type):

        def create_ts_list(dfn):
            cycle_list = dfn.cycle.unique()
            ts_list = []
            for cycle_n in cycle_list:
                ts = dfn[dfn.cycle == cycle_n].value
                ts_list.append(ts)
            return ts_list

        if barycenter_type == 'euc':
            br = list(itertools.chain.from_iterable(tslearn.barycenters.euclidean_barycenter(create_ts_list(df_tag))))
        elif barycenter_type == 'dtw_avg':
            br = list(itertools.chain.from_iterable(tslearn.barycenters.dtw_barycenter_averaging(create_ts_list(df_tag),max_iter=50, tol=1e-3)))
        elif barycenter_type == 'dtw_avg_gradient':
            br = list(itertools.chain.from_iterable(tslearn.barycenters.dtw_barycenter_averaging_subgradient(create_ts_list(df_tag),max_iter=10,random_state=0)))
        elif barycenter_type == 'sdtw':
            br = list(itertools.chain.from_iterable(tslearn.barycenters.softdtw_barycenter(create_ts_list(df_tag), max_iter=5)))
        return br

    def calculate_single_golden_mp(self, df_tag, distance_metric, return_dict, real, barycenter_type):
        tagn = df_tag.tag.unique()[0]
        cycle_cnt = len(df_tag.cycle.unique())
        logger.info("calculating barycenter for tag {} ({} cycles)".format(tagn, cycle_cnt))
        br = self.get_barycenter(df_tag,barycenter_type)
        if real:
            logger.info("evaluating closest real cycle")
            br = self.get_closest_real_cycle(br, df_tag, distance_metric)
            br = br.reshape(-1).tolist()
        return_dict[tagn] = br

    def compute_golden(self,df, real, distance_metric, barycenter_type):
        jobs = []
        manager = mp.Manager()
        return_dict = manager.dict()
        for tagname in df.tag.unique():
            df_tag = df[df.tag == tagname]
            p = mp.Process(target=self.calculate_single_golden_mp, args=(df_tag, distance_metric, return_dict, real, barycenter_type))
            jobs.append(p)
            p.start()   
        for p in jobs:
            p.join()
        return return_dict
   

def real_golden_cycle(df, distance_metric = [], barycenter_type = 'dtw_avg'):

    golden_class = GoldenCycle()
    golden_cycle = golden_class.compute_golden(df, distance_metric = distance_metric, barycenter_type = barycenter_type, real = True)
    return golden_cycle.copy()


def nonreal_golden_cycle(df, barycenter_type = 'dtw_avg'):
    
    distance_metric = None
    golden_class = GoldenCycle()
    golden_cycle = golden_class.compute_golden(df, distance_metric = distance_metric, barycenter_type = barycenter_type, real = False)
    return golden_cycle.copy()


"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

import numpy as np
import pandas as pd
from aimltools.ts.distances import tsdist
from aimltools.ts.denoising import denoise
import tslearn.barycenters
from scipy import stats
import logging
from deprecation import deprecated
from abc import ABCMeta, abstractmethod
from PyNomaly import loop


from pyod.models.lscp import LSCP
from pyod.models.lof import LOF
from pyod.models.feature_bagging import FeatureBagging
from pyod.models.pca import PCA
from pyod.models.auto_encoder import AutoEncoder
from keras import backend as K

logger = logging.getLogger(__name__)


class ClassifierBaseMixin(metaclass = ABCMeta):

    """
    Mixin class containing functionality specific to
    classifiers

    Also contains abstract methods
    """
    
    @abstractmethod
    def fit(self, X, y):
        """
        Method to fit a model from training data
        Mandatorily implement this method in all child mixins
        """
        
    @abstractmethod
    def predict(self, X_new):
        """
        Method to predict labels after classification on new data 
        from a trained model
        Mandatorily implement this method in all child mixins
        """
        
    @abstractmethod
    def predict_proba(self, X_new):
        """
        Method to predict probability scores of classification on new data 
        from a trained model
        Mandatorily implement this method in all child mixins"""


class LOOPClassifier(ClassifierBaseMixin):
    
    """
    Class that implements classification based on the Local Outlier Probabilities (LoOP)
    method, the implementation of which is taken from the PyNomaly library - 
    (https://github.com/vc1492a/PyNomaly)
    """

    def __repr__(self):
        return """LOOP classifier\n
            Arguments: n_neighbors={}, threshold={}, use_numba={}
        """.format(self.__n_neighbors, self.__threshold, self.__use_numba)
    
    def __init__(self, n_neighbors=10, use_numba=True, threshold = 0.5):
        self.__n_neighbors = n_neighbors
        self.__use_numba = use_numba
        self.__threshold = threshold    
    
    def fit(self, X):
        try:
            self.model = loop.LocalOutlierProbability(X, n_neighbors=self.__n_neighbors, use_numba=self.__use_numba)        
            self.model.fit()
        except Exception as error:
            logger.error("Error while training using PyNomaly.loop: {}".format(error))

    def predict(self, X):
        try:
            probs = self.predict_proba(X)
            preds = []
            for p in probs:
                res = 1 if p > self.__threshold else 0
                preds.append(res)
            return np.array(preds)
        except Exception as error:
            logger.error("Error while prediction using PyNomaly.loop: {}".format(error))
    
    def predict_proba(self, X):
        try:
            probs = []
            if (type(X) is pd.core.frame.DataFrame): 
                for itpl in X.itertuples():
                    score = self.model.stream(np.array(itpl)[1:])
                    probs.append(score)
            elif (type(X) is np.ndarray):
                for i in range(0,len(X)):
                    score = self.model.stream(X[i])
                    probs.append(score)
                    
            return np.array(probs)
        except Exception as error:
            logger.error("Error while predicting probabilities using PyNomaly.loop: {}".format(error))
            

class PyODClassifier(ClassifierBaseMixin):
    """
    Class that implements classification based on models that are part of the PyOD library
    """
    @abstractmethod
    def __repr__(self):
        """
        Mandatorily implement this method in all child classes
        """
    
    def __init__(self):
        self._model = {"NA"}
        
    @abstractmethod
    def fit(self, X):
        """
        Mandatorily implement this method in all child classes
        """
        
    def predict(self, X):
        try:
            preds = self._model.predict(X)
            return preds
        except Exception as error:
            logger.error("Error while predicting using {} ({})".format(self._model, error))
            
    def predict_proba(self, X):
        try:
            probs = self._model.predict_proba(X)[:,1]
            return probs 
        except Exception as error:
            logger.error("Error while predicting probabilities using {} ({})".format(self._model, error))
            

class ANNClassifier(PyODClassifier):
    """
    Class that implements classification based on AutoEncoder NN
    method, the implementation of which is taken from the PyOD library
    """
    
    def __repr__(self):
        return "AutoEncoder NN classifier.\n{}".format(self._model)
    
    def __init__(self, outliers_fraction = 0.001, epochs = 30, hidden_neurons = [], reset_keras_before_fit = True):
        super().__init__()
        self.__outliers_fraction = outliers_fraction
        self.__epochs = epochs
        self.__hidden_neurons = hidden_neurons
        self.__reset_keras_before_fit = reset_keras_before_fit
        self.__random_state = None
    
    def set_hidden_neurons(self, hidden_neurons):
        self.__hidden_neurons = hidden_neurons

    def reset_keras_before_fit(self, reset_keras_after_fit):
        self.__reset_keras_before_fit = reset_keras_after_fit

    def set_random_state(self, random_state):
        self.__random_state = random_state

    def fit(self, X):
        if len(self.__hidden_neurons) == 0:
            try:
                self.__hidden_neurons = [len(X.columns)-1]
            except Exception as error:
                self.__hidden_neurons = [len(X[0])]
        
        try:
            if self.__reset_keras_before_fit:
                K.clear_session()
            self._model = AutoEncoder(epochs=self.__epochs, hidden_neurons=self.__hidden_neurons, contamination=self.__outliers_fraction, verbose = 0, random_state=self.__random_state)
            self._model.fit(X)
        except Exception as error:
            logger.error("Error while training using PyOD.AutoEncoder: {}".format(error))
            
            
class LSCPClassifier(PyODClassifier):
    """
    Class that implements classification based on the Locally Selective Combination of Parallel Outlier Ensembles
    method, the implementation of which is taken from the PyOD library
    """

    def __repr__(self):
        return "LSCP classifier.\n{}".format(self._model)
    
    def __init__(self, threshold = 0.5, detector_list = [LOF(n_neighbors=5), LOF(n_neighbors=10), LOF(n_neighbors=15),
                 LOF(n_neighbors=20), LOF(n_neighbors=25), LOF(n_neighbors=30),
                 LOF(n_neighbors=35), LOF(n_neighbors=40), LOF(n_neighbors=45),
                 LOF(n_neighbors=50)]):
        super().__init__()
        self.__threshold = threshold
        self.__detector_list = detector_list
        
    def fit(self, X):
        try:
            self._model = LSCP(self.__detector_list, random_state=42)
            self._model.fit(X)
        except Exception as error:
            logger.error("Error while training with PyOD.LSCPClassifier: {}".format(error))
            

class FeatureBaggingClassifier(PyODClassifier):
    """
    Class that implements classification based on Feature bagging
    method, the implementation of which is taken from the PyOD library
    """
    
    def __repr__(self):
        return "Feature bagging classifier.\n{}".format(self._model)
    
    def __init__(self, check_estimator=False):
        super().__init__()
        self.__check_estimator = check_estimator
    
    def fit(self, X):
        try:
            self._model = FeatureBagging(check_estimator = self.__check_estimator)
            self._model.fit(X)
        except Exception as error:
            logger.error("Error while training with PyOD.FeatureBaggingClassifier: {}".format(error))
            

class PCAClassifier(PyODClassifier):
    """
    Class that implements classification based on Principal Component Analysis (PCA)
    method, the implementation of which is taken from the PyOD library
    """
    
    def __repr__(self):
        return "Principal Component Analysis (PCA) classifier.\n{}".format(self._model)
    
    def __init__(self, outliers_fraction = 0.001):
        super().__init__()
        self.__outliers_fraction = outliers_fraction
    
    def fit(self, X):
        try:
            self._model = PCA(contamination=self.__outliers_fraction, random_state=1)
            self._model.fit(X)
        except Exception as error:
            logger.error("Error while training with PyOD.PCA: {}".format(error))
            

@deprecated(deprecated_in="0.7.0", removed_in="0.8.0")
class PercentileClassifier(ClassifierBaseMixin):

    """
    Class that implements classification using an 
    arbitrarily defined percentile value
    """

    def fit(self, X, cutoff_quantile, tag, reference_cycle, distance_type):
    
        self.tag = tag
        self.reference_cycle = reference_cycle
        self.distance_type = distance_type
        
        self.ref_ts = np.array(X[(X.tag == self.tag) & (X.cycle == self.reference_cycle)].value)
        df_ts = X[(X.tag == self.tag) & (X.cycle != self.reference_cycle)]
        self.distances = calculate_distances(df_ts, self.ref_ts, distance_type)
        self.quant_value = np.quantile(list(self.distances.values()), cutoff_quantile)
    
    
    def predict(self, X):
        
        df = X[(X.tag == self.tag)]    
        self.distances = calculate_distances(df, self.ref_ts, distance_type=self.distance_type)
        self.cycles = []
        self.pred_labels = []
    
        for key, value in self.distances.items():
            self.cycles.append(key)
            if value > self.quant_value:
                self.pred_labels.append(1)
            else:
                self.pred_labels.append(0)
                
        return self.pred_labels
    
    def predict_proba(self, X):
        raise NotImplementedError("Model is not probabilistic")

@deprecated(deprecated_in="0.7.0", removed_in="0.8.0")        
class KDEClassifier(ClassifierBaseMixin):
    
    """
    Class that implements classification using an 
    kernel density estimation on training data and an
    arbitrary threshold value
    """

    def fit(self, X, threshold, tag, reference_cycle, distance_type):
        
        self.tag = tag
        self.reference_cycle = reference_cycle
        self.distance_type = distance_type
        self.threshold = 1 - threshold
        
        self.ref_ts = np.array(X[(X.tag == self.tag) & (X.cycle == self.reference_cycle)].value)
        df_ts = X[(X.tag == self.tag) & (X.cycle != self.reference_cycle)]
        distances = calculate_distances(df_ts, self.ref_ts, distance_type)
        
        self.sample_pdf = stats.gaussian_kde(list(distances.values()))

        
    def predict(self, X):
    
        df = X[(X.tag == self.tag)]
        self.distances = calculate_distances(df, self.ref_ts, distance_type=self.distance_type)
        self.cycles = []
        self.probs = []
        self.pred_labels = []
        
        for key, value in self.distances.items():
            
            self.cycles.append(key)
            prob = self.sample_pdf.integrate_box_1d(value, max(self.distances.values()))
            self.probs.append(prob)
            
            if prob < self.threshold:
                self.pred_labels.append(1)
            else:
                self.pred_labels.append(0)
                
        return self.pred_labels
    
    def predict_proba(self, X):
        return self.probs
    
@deprecated(deprecated_in="0.7.0", removed_in="0.8.0")
class ClassifierBaseClass(metaclass = ABCMeta):

    def __init__(self, X, classifier, **kwargs):
        self.X = X
        self.classifier = classifier
        self.kwargs = kwargs
    
    def __repr__(self):
        return """Classification method = {}, \n
            Arguments = {}""".format(self.classifier, self.kwargs)

    @abstractmethod
    def classify(self):
        """
        Mandatorily implement this method in all child classes
        """

@deprecated(deprecated_in="0.7.0", removed_in="0.8.0")
def calculate_distances(df_ts, ref_ts, distance_type="sdtw"):

    """
    Function to calculate distances between a reference signal
    and multiple other signals

    Documentation to be written
    """

    cycle_list = df_ts.cycle.unique()
    distances_dict = {}
    for cycle_n in cycle_list:
        ts = df_ts[df_ts.cycle == cycle_n].value
        d = tsdist(ts, ref_ts, distance_type=distance_type)
        distances_dict[cycle_n] = d
    return distances_dict
    
@deprecated(deprecated_in="0.7.0", removed_in="0.8.0")
def calculate_barycenter(df_ts, apply_denoise = False):

    """
    Function to compute barycenter of multiple signals

    Documentation to be written
    """

    cycle_list = df_ts.cycle.unique()
    ts_list = []
    for cycle_n in cycle_list:
        ts = df_ts[df_ts.cycle == cycle_n].value
        if apply_denoise == True:
            ts = denoise(ts, denoise_type='wavelet', level = None, 
            threshold = 0.7, axis = -1, wavedec_mode = 'symmetric', 
            return_noise=False, threshold_mode='soft')
        ts_list.append(ts)
    return tslearn.barycenters.dtw_barycenter_averaging(ts_list, max_iter=50, tol=1e-3)

@deprecated(deprecated_in="0.7.0", removed_in="0.8.0")
def insert_cycle(df_ts, value_new_cycle, tag):

    """
    Function to insert a new cycle into a dataframe containing
    data on existing cycles

    Documentation to be written
    """
    
    ts = time.gmtime()
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S", ts)
    
    value = value_new_cycle
    cycles = df_ts.cycle
    new_cycle_num = max(cycles) + 1
    
    assert new_cycle_num not in cycles
    
    df_new_cycle = pd.DataFrame(value, columns=["value"])
    
    df_new_cycle['time'] = timestamp
    df_new_cycle['cycle'] = new_cycle_num
    df_new_cycle['tag'] = tag
    
    df_ts = pd.concat([df_ts, df_new_cycle], axis=0, sort=False)
    df_ts.reset_index(inplace=True)
    
    return df_ts


"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

"""
Sub-module of TSAnalysis

Contains classes and methods to induce perturbations on signals
"""

import numpy as np
import pandas as pd
from abc import ABCMeta, abstractmethod
from scipy.stats import norm
from aimltools.ts.denoising import denoise
import logging

logger = logging.getLogger(__name__)


# Defines perturbation mixins
class BaseMixin(metaclass = ABCMeta):
    """
    Defines pattern for child mixins
    """

    @abstractmethod
    def perturb_transform(self):
        """Mandatorily implement this method in all child mixins"""

    @abstractmethod
    def perturb_transform_(self):
        """Mandatorily implement this method in all child mixins"""

class SignalTimeDilationMixin(BaseMixin):
    """
    Mixin class for perturbing signal through signal time dilation
    """
    def __init__(self, D):
        self.__D = D
    
    def perturb_transform(self, signal, D):
        if D==1:
            return signal
        else:
            xp = np.arange(len(signal))
            b = int(D * len(signal))
            signal = np.interp(np.linspace(0,len(signal),b), xp, signal)
            return signal

    def perturb_transform_(self, signal):
        return self.perturb_transform(signal, D = self.__D)

class HorizontalShiftMixin(BaseMixin):
    """
    Mixin class for perturbing signal through horizontal shifting
    """
    def __init__(self, S, method):
        self.__S = S
        self.__method = method

    def perturb_transform(self, signal, S, method = 'proportional'):
        
        if method == 'proportional':
            step = round(len(signal) * S)
            
            d = signal[0] - signal[-1]
            start = signal[-step:] + d
            signal = np.concatenate([start,signal[:-step]])
            return signal
        
        if method == 'absolute':
            step = S
            
        d = signal[0] - signal[-1]
        start = signal[-step:] + d
        signal = np.concatenate([start,signal[:-step]])
        return signal

    def perturb_transform_(self, signal):
        return self.perturb_transform(signal, S = self.__S, method = self.__method)

class VerticalShiftMixin(BaseMixin):
    """
    Mixin class for perturbing signal through vertical shifting
    """
    def __init__(self, S, method):
        self.__S = S
        self.__method = method

    def perturb_transform(self, signal, S, method = 'proportional'):
        
        if method == 'proportional':
            amplitude_range = abs(max(signal) - min(signal))
            step = amplitude_range * S
            
        if method == 'absolute':
            step = S
            
        signal = signal + step
        
        return signal

    def perturb_transform_(self, signal):
        return self.perturb_transform(signal, S = self.__S, 
                                        method = self.__method)

class AmplitudeVariationMixin(BaseMixin):
    """
    Mixin class for perturbing signal through amplitude variation
    """
    def __init__(self, A):
        self.__A = A

    def perturb_transform(self, signal, A):
        # Denoise through DWT
        signal, noise = denoise(signal, return_noise=True)
        signal = signal*A + noise
        return signal

    def perturb_transform_(self, signal):
        return self.perturb_transform(signal, A = self.__A)

class GaussianPeakMixin(BaseMixin):
    """
    Mixin class for perturbing signal through Gaussian kernel
    """
    def __init__(self, h, n_peaks="random"):
        self.__h = h
        self.__n_peaks = n_peaks

    def perturb_transform(self, signal, h, n_peaks="random"):
        x_all = np.arange(0, signal.shape[0], 1) 
        A = (max(signal) - min(signal))
        if n_peaks == "random":
            n_peaks = np.random.choice(np.arange(1,6))
        
        for _ in range(n_peaks):
            mean = round(np.random.normal(loc=signal.shape[0]/2, scale=signal.shape[0]/8))
            std = np.random.choice(np.arange(1,4, step=0.1), replace=True)
            y2 = norm.pdf(x_all, mean, std)
            try:
                if signal[mean] < 0:
                    signal = signal - y2*A*h
                else:
                    signal = signal + y2*A*h
            except:
                signal = signal + y2*A*h
            
        return signal

    def perturb_transform_(self, signal):
        return self.perturb_transform(signal, h = self.__h, n_peaks=self.__n_peaks)


class SimplePeakMixin(BaseMixin):
    def __init__(self, h):
        self.__h = h

    def perturb_transform(self, signal, h):
        x_all = np.arange(0, signal.shape[0], 1)
        A = (max(signal) - min(signal))
        y2 = norm.pdf(x_all,signal.shape[0]/2,2)
        signal = signal + y2*A*h
        return signal

    def perturb_transform_(self, signal):
        return self.perturb_transform(signal, h = self.__h)

# Defines base class
class PerturbationBaseClass(metaclass = ABCMeta):
    """"
    Base class containing common functionality and abstract methods
    """
    
    def __init__(self, signal, perturbation_type, **kwargs):
        self.signal = signal
        self.perturbation_type = perturbation_type
        self.kwargs = kwargs

    def __repr__(self):
        return """Perturbation type = {}, \n
                Arguments = {}""".format(self.perturbation_type, self.kwargs)
    
    @abstractmethod
    def perturb_method(self):
        """Mandatorily implement this method in all child classes"""


# Defines user interface   
def perturb(signal, perturbation_type = 'horizontal', **kwargs):

    """
    Function that returns an input signal after inducing a perturbation

    Parameters
    ----------
    signal : numpy.ndarray
                The input signal

    perturbation_type : str
                Type of perturbation to be induced. Defaults to 'horizontal'
                Currently supported perturbations are:
                'horizontal' = Horizontal shift
                'vertical' = Vertical shift
                'amplitude' = Amplitude variation
                'dilation' = Signal time dilation
                'gaussian' = Gaussian peak
                'peak' = Simple peak at the center of the signal
    
    kwargs: dict
                Keyword arguments to be passed depending on value of 
                'perturbation_type'

    Returns
    -------
    signal_perturbed : numpy.ndarray
                The perturbed signal

    """
    
    # Contains the perturbation_type and its corresponding mixin
    # Add a new {perturbation_type: mixin} key-value pair to 
    # register a new perturbation method
    perturb_mixin_match_dict = {
                                'horizontal': HorizontalShiftMixin,
                                'vertical': VerticalShiftMixin,
                                'dilation': SignalTimeDilationMixin,
                                'amplitude': AmplitudeVariationMixin,
                                'gaussian': GaussianPeakMixin,
                                'peak': SimplePeakMixin
                              }

    supported = list(perturb_mixin_match_dict.keys())
    
    try:
        mixin_to_use = perturb_mixin_match_dict[perturbation_type]
    except:
        raise ValueError("""
        Unsupported value '{}' for perturbation_type. 
        Currently supported perturbations are -
        {}""".format(perturbation_type, supported))

    # Throws exception if mixin_to_use is not a child of BaseMixin
    assert issubclass(mixin_to_use, BaseMixin)

    # Defines child class from base class and required mixin
    class PerturbationSimulator(PerturbationBaseClass, mixin_to_use):

        def __init__(self, signal, perturbation_type, **kwargs):
            super().__init__(signal, perturbation_type, **kwargs)
                
        def perturb_method(self):
            signal_perturbed = self.perturb_transform(signal, **self.kwargs)
            return signal_perturbed
    
    # Instatiates child class
    instance = PerturbationSimulator(signal, perturbation_type, **kwargs)
    signal_perturbed = instance.perturb_method()
    return signal_perturbed


class CyclePerturbationHelper:
    """
    Helper class that's used to apply a perturbation to the signal column in a data frame having specific structure.
    Perturbation is always returned in a new data frame. Original data frame is left untouched.
    Input data frame should have the following columns:
        - tag (string): name of the variable being sampled
        - time (numpy.datetime64): time at which the variable has been sampled
        - value (numpy.float64): sampled value
        - cycle (numpy.int64): used to group sampled measurement into sets (e.g.: group variable
            measurements collected during the execution of a process that's repeated on a regular basis)
    """
    def __init__(self, tag_col, time_col, signal_col, cycle_col, tag_list = []):
        """

        :param tags_col: name of tag column
        :param time_col: name of time column
        :param signal_col: name of signal value column
        :param cycle_col: name of cycle column
        :param tag_list: variables to which the perturbation should be applied
        """
        self.__tag_col = tag_col
        self.__time_col = time_col
        self.__signal_col = signal_col
        self.__cycle_col = cycle_col
        self.__tag_list = tag_list

    def set_tag_list(self, tag_list):
        self.__tag_list = tag_list

    def set_perturbation(self, pmixin_class):
        """

        :param pmixin_class: instance of a valid perturbation class
        :return: NA
        """
        assert (isinstance(pmixin_class, BaseMixin))
        self.__pmixin_class = pmixin_class

    def perturb_cycles(self, df):
        """
        Applies a given perturbation to the given data frame

        :param df: pandas.DataFrame
                        The input data frame
        :return df_p : pandas.DataFrame
                        The perturbed data frame
        """

        cycles_list = df[self.__cycle_col].unique()
        if len(self.__tag_list) == 0:
            tags_list_new = df[self.__tag_col].unique()
        else:
            tags_list_new = list(set(df[self.__tag_col].unique()) & set(self.__tag_list))

        if isinstance(self.__pmixin_class, SignalTimeDilationMixin):
            df_p = self.__perturb_cycles_dilation(df, cycles_list, tags_list_new)
        else:
            df_p = self.__perturb_cycles_generic(df, cycles_list, tags_list_new)
        return df_p

    def __perturb_cycles_generic(self, df, cycle_list, tag_list):
        df_test = df.copy()
        for cycle in cycle_list:
            df_cycle = df_test[df_test[self.__cycle_col] == cycle]
            for tag in tag_list:
                idxvals = df_cycle[df_cycle[self.__tag_col] == tag].index
                if len(idxvals) == 0:
                    continue
                tmp_arr = df_cycle.loc[idxvals, (self.__signal_col)].values
                tmp_arr = self.__pmixin_class.perturb_transform_(tmp_arr)
                df_test.loc[idxvals, (self.__signal_col)] = tmp_arr
        return df_test

    def __perturb_cycles_dilation(self, df, cycle_list, tag_list):
        tag_lst = []
        time_lst = []
        value_lst = []
        cycle_lst = []
        for cycle in cycle_list:
            df_cycle = df[df[self.__cycle_col] == cycle]
            for tag in tag_list:
                if len(df_cycle[df_cycle[self.__tag_col] == tag]) == 0:
                    continue
                [tag_l, time_l, value_l, cycle_l] = self.__apply_dilation(df_cycle[df_cycle[self.__tag_col] == tag])
                tag_lst.extend(tag_l)
                time_lst.extend(time_l)
                value_lst.extend(value_l)
                cycle_lst.extend(cycle_l)
        df_p = pd.DataFrame(list(zip(tag_lst, time_lst, value_lst, cycle_lst)), columns=df.columns)
        return df_p

    def __apply_dilation(self, df):
        assert (len(df.cycle.unique()) <= 1)
        assert (len(df.tag.unique()) <= 1)
        unit = np.datetime_data(df[self.__time_col].values[0])[0]
        # apply dilation to time column
        time_d = self.__pmixin_class.perturb_transform_(df[self.__time_col].values.astype('int'))
        time_d = [np.datetime64(int(t), unit) for t in time_d]
        # apply dilation to signal column
        values_d = self.__pmixin_class.perturb_transform_(df[self.__signal_col].values)
        # re-sample tag column and cycle column
        tags_d = [df.tag.unique()[0]] * len(time_d)
        cycles_d = [df.cycle.unique()[0]] * len(time_d)
        return [tags_d, time_d, values_d, cycles_d]


import sys
import clr

OSI_AFSDK_PATH = r'C:\Program Files (x86)\PIPC\AF\PublicAssemblies\4.0'
OSI_PISDK_PATH = r'C:\Program Files\PIPC\pisdk\PublicAssemblies'
LLY_PIDRSDK_PATH = r'C:\Program Files\Eli Lilly\PIDataReaderApps'
MS_DOTNET_PATH = r'C:\Program Files (x86)\Reference Assemblies\Microsoft\Framework\.NETFramework\v4.6.2'

sys.path.append(OSI_AFSDK_PATH)
sys.path.append(OSI_PISDK_PATH)
sys.path.append(LLY_PIDRSDK_PATH)
sys.path.append(MS_DOTNET_PATH)

clr.AddReference('OSIsoft.AFSDK')
clr.AddReference('OSIsoft.PISDK')
clr.AddReference("System.Collections")
clr.AddReference('PIDataReaderLib')
clr.AddReference('mscorlib')

from OSIsoft.AF import *
from OSIsoft.AF.PI import *
from OSIsoft.AF.Asset import *
from OSIsoft.AF.Data import *
from OSIsoft.AF.Time import *
from OSIsoft.AF.UnitsOfMeasure import *
from System import *
from System.Collections.Generic import List
from System import DateTime, DateTimeOffset
from PIDataReaderLib import *


"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

"""
Sub-module of TSAnalysis

Contains classes and methods to compute distances between signals
"""

import numpy as np
from tslearn.metrics import soft_dtw, dtw
from aimltools.ts.dependencies import PD_distance
from aimltools.ts.dependencies import tam_distance as tam
from abc import ABCMeta, abstractmethod
import logging

logger = logging.getLogger(__name__)


# Defines distance mixins
class BaseMixin(metaclass = ABCMeta):
    """
    Defines pattern for child mixins
    """

    @abstractmethod
    def compute_distance(self):
        """Mandatorily implement this method in all child mixins"""

    @abstractmethod
    def compute_distance_(self):
        """Mandatorily implement this method in all child mixins"""


class EuclideanMixin(BaseMixin): 
    """
    Mixin class for computing Euclidean distance between two signals
    """

    def compute_distance(self, signalA, signalB):
        
        # Checks if the arrays are of equal lengths
        try:
            assert len(signalA) == len(signalB)
        except:
            raise AssertionError("""
            Expected arrays of equal lengths. 
            Received lengths of {} and {} respectively. 
            """.format(len(signalA), len(signalB)))
        
        euclidean_dist = np.linalg.norm(signalA - signalB, ord=2)
        return euclidean_dist

    def compute_distance_(self, signalA, signalB):
        return self.compute_distance(signalA, signalB)


class PDMixin(BaseMixin):  
    """
    Mixin class for computing Permutation Distance between two signals
    """ 
    def __init__(self, metric = 'kullback', m = 3, step = 1):
        self.__metric = metric
        self.__m = m
        self.__step = step

    def compute_distance(self, signalA, signalB, metric = 'kullback', m = 3, step = 1):
        metric = metric.lower()
        logger.info("Divergence metric set to {}".format(metric.title()))
        pd_distance = PD_distance(signalA, signalB, metric, m, step)
        return pd_distance

    def compute_distance_(self, signalA, signalB):
        return self.compute_distance(signalA, signalB, metric = self.__metric, m = self.__m, step = self.__step)


class SoftDTWMixin(BaseMixin):
    """
    Mixin class for computing Soft Dynamic Time Warping distance between two signals
    """
    def __init__(self, gamma=1.0, normalize = True):
        self.__gamma = gamma
        self.__normalize = normalize

    def compute_distance(self, signalA, signalB, gamma=1.0, normalize = True):
        if normalize == True:
            soft_dtw_distance_AB = soft_dtw(signalA, signalB, gamma)
            soft_dtw_distance_AA = soft_dtw(signalA, signalA, gamma)
            soft_dtw_distance_BB = soft_dtw(signalB, signalB, gamma)
            soft_dtw_distance = soft_dtw_distance_AB - (1 / 2) * ((soft_dtw_distance_AA) + (soft_dtw_distance_BB))
        else:
            soft_dtw_distance = soft_dtw(signalA, signalB, gamma)

        return soft_dtw_distance

    def compute_distance_(self, signalA, signalB):
        return self.compute_distance(signalA, signalB, gamma=self.__gamma, normalize=self.__normalize)


class DTWMixin(BaseMixin):
    """
    Mixin class for computing Dynamic Time Warping distance between two signals
    """
    def __init__(self, global_constraint=None, sakoe_chiba_radius=None, itakura_max_slope=None):
        self.__global_constraint = global_constraint
        self.__sakoe_chiba_radius = sakoe_chiba_radius
        self.__itakura_max_slope = itakura_max_slope

    def compute_distance(self, signalA, signalB, global_constraint=None, 
                            sakoe_chiba_radius=None, itakura_max_slope=None):
        dtw_distance = dtw(signalA, signalB, global_constraint, 
                            sakoe_chiba_radius, itakura_max_slope)
        return dtw_distance

    def compute_distance_(self, signalA, signalB):
        return self.compute_distance(signalA, signalB, global_constraint=self.__global_constraint, sakoe_chiba_radius=self.__sakoe_chiba_radius, itakura_max_slope=self.__itakura_max_slope)

class TAMMixin(BaseMixin):
    """
    Mixin class for computing Time Alignment Measure distance between two signals 
    """ 

    def compute_distance(self, signalA, signalB):
        tam_dist = tam(signalA, signalB)
        return tam_dist

    def compute_distance_(self, signalA, signalB):
        return self.compute_distance(signalA, signalB)

# Defines base class
class TSDistanceBaseClass(metaclass = ABCMeta):
    """
    Base class containing common functionality and abstract methods
    """

    def __init__(self, signalA, signalB, distance_type, **kwargs):
        self.signalA = signalA
        self.signalB = signalB
        self.distance_type = distance_type
        self.kwargs = kwargs

    def __repr__(self):
        return """Distance type = {}, \n
                Arguments = {}""".format(self.distance_type, self.kwargs)
    
    @abstractmethod
    def distance_method(self):
        """Mandatorily implement this method in all child classes"""


# Defines user interface
def tsdist(signalA, signalB, distance_type = 'euclidean',
            **kwargs):

    """
    A function to compute the distance between two time signal arrays

    Parameters
    ----------
    signalA : np.ndarray
            An array containing a signal/time series

    signalB : np.ndarray
            Another array containing a signal/time series

    distance_type : str
            The distance measure used for computation. Defaults to 'euclidean'
            Supported distances are:
            'euclidean' = Euclidean distance
            'pdc' = Permutation distribution clustering
            'tam' = Time alignment measure
            'dtw' = Dynamic time warping
            'sdtw' = Short dynamic time warping
            
    kwargs : dict
            Keyword arguments to be passed depending on
            the value of 'type'

    Returns
    -------
    distance : np.float64
            The computed distance metric
    """

    # Dict of a distance_type and its corresponding mixin
    # Add a new {distance_type: mixin} key-value pair to 
    # register a new distance implementation 
    distance_mixin_match_dict = {
                                'euclidean': EuclideanMixin,
                                'pdc': PDMixin,
                                'tam': TAMMixin,
                                'dtw': DTWMixin,
                                'sdtw': SoftDTWMixin
                                }

    supported = list(distance_mixin_match_dict.keys())

    try:
        mixin_to_use = distance_mixin_match_dict[distance_type]
    except:
        raise ValueError("""
        Unsupported value '{}' for distance_type. 
        Currently supported distances are -
        {}""".format(distance_type, supported))

    # Throws exception if mixin_to_use is not a child of BaseMixin
    assert issubclass(mixin_to_use, BaseMixin)
                   
    # Defines child class from base class and required mixin
    class TSDistance(TSDistanceBaseClass, mixin_to_use):

        def __init__(self, signalA, signalB, distance_type, **kwargs):
            super().__init__(signalA, signalB, distance_type, **kwargs)
        
        def distance_method(self):
            logging.info("Distance type = {}".format(self.distance_type))
            distance = self.compute_distance(signalA, signalB, **self.kwargs)
            return distance

    # Instantiates child class
    instance = TSDistance(signalA, signalB, distance_type, **kwargs)
    distance = instance.distance_method()
    return distance
            


from enum import Enum
import pandas as pd
import pathlib
import numbers
from pathlib import Path
from seeq import spy
import aimltools

#launch configuration file
p2 = pathlib.PurePath(Path(globals()['__file__']).parent, 'pi_utils_cfg.py')
try:
    with open(p2) as fd:
        exec(fd.read())
except FileNotFoundError:
    print("Module \'{}\' requires to be initialized with a valid configuration file, but no file was found.".format(Path(globals()['__file__'])))
    print("Please ensure that a configuration file named \'{}\' exists in this folder: \'{}\'.".format("pi_utils_cfg.py", Path(globals()['__file__']).parent))
    print("Please ensure that the file contains valid paths to.NET assemblies used for reading data from OSI PI.")
    print("See file \'{}\' for an example.".format("pi_utils_cfg_sample.py"))
    print("Execution will now stop.")
    quit()
except Exception as e:
    print("Unexpected exception importing {}".format(p2))

class BoundaryType(Enum):
    INSIDE = 1
    OUTSIDE = 2
    INTERPOLATED = 3


class PITagValues:

    def __init__(self, name, values):
        self.name = name
        self.values = values


class PITagValue:

    def __init__(self, utc_milliseconds, value):
        self.utc_milliseconds = utc_milliseconds
        self.value = value


class PIReader:
    def read_tags(self, tagname_list, starttime, endtime) -> pd.DataFrame:
        return None

    @staticmethod
    def build_seeq_pireader(url: str, username: str, password: str, proxy: str) -> 'PIReader':
        pi_reader = SeekPIReader(url, username, password, proxy)
        return pi_reader

    @staticmethod
    def build_afsdk_pireader(servername: str, datetime_format: str, datetime_format_pi: str,
                             boundary_type: BoundaryType) -> 'PIReader':
        pi_reader = OSIPIReader(servername, datetime_format, datetime_format_pi, boundary_type)
        return pi_reader


class SeekPIReader(PIReader):

    def __init__(self, url, username, password, proxy):
        spy.login(url=url, username=username, password=password, proxy=proxy)
        self.__max_tags_per_pull = 20
        self.__is_quiet = True
        return

    @property
    def quiet(self):
        return self.__is_quiet

    @quiet.setter
    def quiet(self, q: bool):
        self.__is_quiet = q

    @property
    def max_tags_per_pull(self):
        return self.__max_tags_per_pull

    @max_tags_per_pull.setter
    def max_tags_per_pull(self, max_tags_per_pull):
        self.__max_tags_per_pull = max_tags_per_pull

    def read_capsules(self, capsule_id_list, starttime, endtime, workbook_id = None):
        items = pd.DataFrame()
        items['ID'] = capsule_id_list
        if workbook_id:
            df_ = spy.search(items, workbook=workbook_id)
        else:
            df_ = spy.search(items)
        df_capsules = spy.pull(df_, grid=None, start=starttime, end=endtime, quiet=self.__is_quiet)
        return df_capsules

    def read_tags(self, tagname_list, starttime, endtime):
        df_tags = pd.DataFrame()
        while len(tagname_list) > 0:
            taglist_ = tagname_list[:self.__max_tags_per_pull]
            df_unpivoted = self.__pull_tags(taglist_, starttime, endtime)
            df_tags = df_tags.append(df_unpivoted)
            tagname_list = tagname_list[self.__max_tags_per_pull:]

        df_tags['time'] = (df_tags['time'].astype('int64')/1e+06).astype('int64')
        df_tags.sort_values(by='time', inplace=True)
        df_tags.reset_index(drop=True, inplace=True)
        return df_tags

    def __pull_tags(self, tagname_list, starttime, endtime):
        items = pd.DataFrame()
        items['Name'] = tagname_list
        df_ = spy.search(items)
        df = spy.pull(df_, grid=None, start=starttime, end=endtime, quiet=self.__is_quiet)
        df_unpivoted = pd.melt(df.reset_index(), id_vars='index', var_name="tag", value_name="value")
        df_unpivoted.rename(columns={"index": "time"}, inplace=True)
        df_unpivoted.dropna(inplace=True)
        #reorder columns
        df_unpivoted = df_unpivoted[['tag', 'time', 'value']]
        return df_unpivoted

class OSIPIReader(PIReader):

    def __init__(self, servername, datetime_format, datetime_format_pi, boundary_type = BoundaryType.INSIDE):
        self.piServers = PIServers()
        self.servername = servername
        self.piServer = self.piServers[servername]
        self.datetime_format = datetime_format
        self.datetime_format_pi = datetime_format_pi
        self.boundary_type = boundary_type

        self.__COLNAME_TAG = 'tag'
        self.__COLNAME_TIME = 'time'
        self.__COLNAME_VALUE = 'value'
        self.__COLNAME_SVALUE = 'svalue'

        self.__df_column_names = [self.__COLNAME_TAG, self.__COLNAME_TIME, self.__COLNAME_VALUE, self.__COLNAME_SVALUE]


    @property
    def datetime_format(self):
        return self.__datetime_format

    @datetime_format.setter
    def datetime_format(self, datetime_format):
        self.__datetime_format = datetime_format

    def read_tags(self, tagname_list, starttime, endtime):
        tagname_list = list(set(tagname_list))
        pi_tag_df = pd.DataFrame(columns=self.__df_column_names)
        for tagname in tagname_list:
            tags_coll = self.__read_tag(tagname, starttime, endtime, self.boundary_type, serialize="df")
            pi_tag_df = pd.concat([pi_tag_df, tags_coll])

        pi_tag_df['time'] = pi_tag_df['time'].astype('int64')
        pi_tag_df.sort_values(by=['time'], ascending=True, inplace=True)
        pi_tag_df.reset_index(drop=True, inplace=True)
        return pi_tag_df

    def read_tags_aslist(self, tagname_list, starttime, endtime):
        tagname_list = list(set(tagname_list))
        pi_tag_values = []
        for tagname in tagname_list:
            tags_coll = self.__read_tag(tagname, starttime, endtime, self.boundary_type, serialize="list")
            pi_tag_values.append(tags_coll)

        return pi_tag_values

    def read_tag_interpolated(self, tagname, ts_list):
        pt = PIPoint.FindPIPoint(self.piServer, tagname)
        pt_list = PIPointList()
        pt_list.Add(pt)

        li = List[AFTime]()
        for ts in ts_list:
            ts_epoch = aimltools.ts.utils.generic_utils.DateTimeUtils.string_to_epoch_ciso8601(ts)
            ts_af = AFTime(ts_epoch)
            li.Add(ts_af)

        listResults = pt_list.InterpolatedValuesAtTimes(li, "", False, PIPagingConfiguration(PIPageType.TagCount, 1))

        pi_tag_value_list = list()
        for afVals in listResults:
            for afVal in afVals:
                dto = DateTimeOffset(afVal.Timestamp.LocalTime)
                pi_tag_value = PITagValue(dto.ToUnixTimeMilliseconds(), afVal.Value)
                pi_tag_value_list.append(pi_tag_value)

        pi_tag_values = PITagValues(tagname, pi_tag_value_list)
        return pi_tag_values

    def read_batch_tree(self, modulepath, starttime, endtime):
        start_time_dt = aimltools.ts.utils.generic_utils.DateTimeUtils.string_to_datetime_ciso8601(starttime)
        end_time_dt = aimltools.ts.utils.generic_utils.DateTimeUtils.string_to_datetime_ciso8601(endtime)

        # generate .Net DateTime
        start_time = DateTime(start_time_dt.year, start_time_dt.month, start_time_dt.day, start_time_dt.hour, start_time_dt.minute, start_time_dt.second)
        end_time = DateTime(end_time_dt.year, end_time_dt.month, end_time_dt.day, end_time_dt.hour, end_time_dt.minute, end_time_dt.second)

        #tzoh = str(start_time_dt.tzinfo)[4:6]
        #tzom = str(start_time_dt.tzinfo)[-2:]
        #tspan = TimeSpan(int(tzoh), int(tzom), 0)
        #start_time = DateTimeOffset(start_time_dt.year, start_time_dt.month, start_time_dt.day, start_time_dt.hour, start_time_dt.minute, start_time_dt.second, tspan)
        #end_time = DateTimeOffset(end_time_dt.year, end_time_dt.month, end_time_dt.day, end_time_dt.hour, end_time_dt.minute, end_time_dt.second, tspan)
        
        reader = PISDKReader(self.servername, self.datetime_format_pi, self.datetime_format_pi, '', '', '')
        pidata = reader.ReadBatchTree(start_time, end_time, modulepath)

        return pidata

    def __read_tag(self, tagname, starttime, endtime, boundary_type, serialize):
        starttime_utc_secs = aimltools.ts.utils.DateTimeUtils.string_to_epoch_ciso8601(starttime)
        endtime_utc_secs = aimltools.ts.utils.DateTimeUtils.string_to_epoch_ciso8601(endtime)

        pt = PIPoint.FindPIPoint(self.piServer, tagname)

        af_starttime = AFTime(starttime_utc_secs)
        af_endtime = AFTime(endtime_utc_secs)

        time_range = OSIPIReader.__setup_time_range_af(af_starttime, af_endtime)

        af_boundary_type = AFBoundaryType.Inside
        if boundary_type == BoundaryType.OUTSIDE:
            af_boundary_type = AFBoundaryType.Outside
        if boundary_type == BoundaryType.INTERPOLATED:
            af_boundary_type = AFBoundaryType.Interpolated

        af_vals = pt.RecordedValues(time_range, af_boundary_type, "", False)

        if len(af_vals) == 0:
            return None

        val_type = OSIPIReader.__getAFValueType(af_vals.get_Item(0))
        if val_type == "unsupported":
            return None

        retval = []
        if serialize == "list":
            pi_tag_value_list = list()
            for afVal in af_vals:
                dto = DateTimeOffset(afVal.Timestamp.LocalTime)

                pi_tag_value = PITagValue(dto.ToUnixTimeMilliseconds(), afVal.Value)
                pi_tag_value_list.append(pi_tag_value)

            pi_tag_values = PITagValues(tagname, pi_tag_value_list)
            retval = pi_tag_values
        elif serialize == "df":
            #serialize in df having columns: tag,time,value (for numerical),svalue (for strings)
            col_tagname = [tagname]*len(af_vals)
            col_timestamp = []
            target_column = []
            for afVal in af_vals:
                dto = DateTimeOffset(afVal.Timestamp.LocalTime)
                col_timestamp.append(dto.ToUnixTimeMilliseconds())
                target_column.append(afVal.Value)

            if val_type == "numeric":
                col_value = target_column
                col_svalue = [""]*len(af_vals)
            else:
                col_svalue = target_column
                col_value = [] * len(af_vals)

            retval = pd.DataFrame(list(zip(col_tagname, col_timestamp, col_value, col_svalue)), columns=self.__df_column_names)

        return retval

    @staticmethod
    def __getAFValueType(afVal):
        v = afVal.Value
        if isinstance(v, numbers.Number):
            return "numeric"
        elif isinstance(v, str):
            return "string"

        return "unsupported"

    @staticmethod
    def __setup_time_range(starttime_str, endtime_str):
        prv = AFTimeZoneFormatProvider(AFTimeZone())
        af_time_range = AFTimeRange(starttime_str, endtime_str, prv)
        return af_time_range

    @staticmethod
    def __setup_time_range_af(starttime_af, endtime_af):
        af_time_range = AFTimeRange(starttime_af, endtime_af)
        return af_time_range


class PIDataUBSerializer:
    COL_UB_UID = 'uid'
    COL_UB_BATCHID = 'batch_id'
    COL_UB_PRODUCT = 'product'
    COL_UB_STARTTIME = 'start_time'
    COL_UB_ENDTIME = 'end_time'
    COL_UB_UTCSTART = 'utc_start'
    COL_UB_UTCEND = 'utc_end'
    COL_UB_PROCEDURE = 'procedure'
    COL_UB_MODULEUID = 'moduleuid'
    COL_UB_BATCHUID = 'batchuid'

    COLS_UB_DATAFRAME = [COL_UB_UID, COL_UB_BATCHID, COL_UB_PRODUCT, COL_UB_STARTTIME, COL_UB_ENDTIME, COL_UB_PROCEDURE, COL_UB_MODULEUID, COL_UB_BATCHUID]

    COL_SB_UID = 'uid'
    COL_SB_UBATCHUID = 'unitbatchuid'
    COL_SB_PATH = 'path'
    COL_SB_NAME = 'name'
    COL_SB_LEVEL = 'level'
    COL_SB_CHLDCNT = 'childcount'
    COL_SB_STARTTIME = 'starttime'
    COL_SB_ENDTIME = 'endtime'
    COL_SB_UTCSTART = 'utcstart'
    COL_SB_UTCEND = 'utcend'
    COL_SB_HEADINGUID = 'headinguid'
    COL_SB_PARENTUID = 'parentuid'

    COLS_SB_DATAFRAME = [COL_SB_UID, COL_SB_UBATCHUID, COL_SB_PATH, COL_SB_NAME, COL_SB_LEVEL, COL_SB_CHLDCNT, COL_SB_STARTTIME, COL_SB_ENDTIME, COL_SB_HEADINGUID, COL_SB_PARENTUID]

    def __init__(self):
        self.__ub_dict = dict.fromkeys(PIDataUBSerializer.COLS_UB_DATAFRAME)
        for key in self.__ub_dict.keys():
            self.__ub_dict[key] = []

        self.__sb_dict = dict.fromkeys(PIDataUBSerializer.COLS_SB_DATAFRAME)
        for key in self.__sb_dict.keys():
            self.__sb_dict[key] = []

    @staticmethod
    def batch_tree_to_ub_sb_df(pidata):
        pidataser = PIDataUBSerializer()
        [df_ub, df_sb] = pidataser.process_batches(pidata)
        return [df_ub, df_sb]

    def process_batches(self, pidata):
        for batch in pidata.get_batches():
            for unit_batch in batch.get_unitBatches():
                self.__process_unit_batch(unit_batch)

        df_ub = pd.DataFrame.from_dict(self.__ub_dict)
        df_ub[PIDataUBSerializer.COL_UB_UTCSTART] = [aimltools.ts.utils.generic_utils.DateTimeUtils.datetime_to_epoch(val) for val in self.__ub_dict[PIDataUBSerializer.COL_UB_STARTTIME]]
        df_ub[PIDataUBSerializer.COL_UB_UTCEND] = [aimltools.ts.utils.generic_utils.DateTimeUtils.datetime_to_epoch(val) for val in self.__ub_dict[PIDataUBSerializer.COL_UB_ENDTIME]]

        df_sb = pd.DataFrame.from_dict(self.__sb_dict)
        df_sb[PIDataUBSerializer.COL_SB_UTCSTART] = [aimltools.ts.utils.generic_utils.DateTimeUtils.datetime_to_epoch(val) for val in self.__sb_dict[PIDataUBSerializer.COL_SB_STARTTIME]]
        df_sb[PIDataUBSerializer.COL_SB_UTCEND] = [aimltools.ts.utils.generic_utils.DateTimeUtils.datetime_to_epoch(val) for val in self.__sb_dict[PIDataUBSerializer.COL_SB_ENDTIME]]

        return [df_ub, df_sb]

    def __process_unit_batch(self, unit_batch):
        self.__ub_dict[PIDataUBSerializer.COL_UB_UID].append(unit_batch.uid)
        self.__ub_dict[PIDataUBSerializer.COL_UB_BATCHID].append(unit_batch.batchid)
        self.__ub_dict[PIDataUBSerializer.COL_UB_PRODUCT].append(unit_batch.product)
        self.__ub_dict[PIDataUBSerializer.COL_UB_STARTTIME].append(pd.Timestamp(unit_batch.starttime, tz='utc'))
        self.__ub_dict[PIDataUBSerializer.COL_UB_ENDTIME].append(pd.Timestamp(unit_batch.endtime, tz='utc'))
        self.__ub_dict[PIDataUBSerializer.COL_UB_PROCEDURE].append(unit_batch.procedure)
        self.__ub_dict[PIDataUBSerializer.COL_UB_MODULEUID].append(unit_batch.moduleuid)
        self.__ub_dict[PIDataUBSerializer.COL_UB_BATCHUID].append(unit_batch.batchuid)
        if len(unit_batch.subBatches) > 0:
            level = 0
            path = "\\"
            self.__process_subbatches("", unit_batch.subBatches, unit_batch, path, level)

    def __process_subbatches(self, parent_uid, sub_batches, reference_unit_batch, path, level):
        for sub_batch in sub_batches:
            self.__process_subbatch(parent_uid, sub_batch, reference_unit_batch, path, level)

    def __process_subbatch(self, parent_uid, sub_batch, reference_unit_batch, path, level):
        self.__sb_dict[PIDataUBSerializer.COL_SB_UID].append(sub_batch.uid)
        self.__sb_dict[PIDataUBSerializer.COL_SB_UBATCHUID].append(reference_unit_batch.uid)
        self.__sb_dict[PIDataUBSerializer.COL_SB_PATH].append(path)
        self.__sb_dict[PIDataUBSerializer.COL_SB_NAME].append(sub_batch.name)
        self.__sb_dict[PIDataUBSerializer.COL_SB_LEVEL].append(level)
        self.__sb_dict[PIDataUBSerializer.COL_SB_CHLDCNT].append(len(sub_batch.subBatches))
        self.__sb_dict[PIDataUBSerializer.COL_SB_STARTTIME].append(pd.Timestamp(sub_batch.starttime, tz='utc'))
        self.__sb_dict[PIDataUBSerializer.COL_SB_ENDTIME].append(pd.Timestamp(sub_batch.endtime, tz='utc'))
        self.__sb_dict[PIDataUBSerializer.COL_SB_HEADINGUID].append(sub_batch.headinguid)
        self.__sb_dict[PIDataUBSerializer.COL_SB_PARENTUID].append(parent_uid)
        self.__iterate_subbatch(sub_batch.uid, sub_batch, reference_unit_batch, path, level)

    def __iterate_subbatch(self, parent_uid, sub_batch, reference_unit_batch, path, level):
        if len(sub_batch.subBatches) == 0:
            return
        path = path + sub_batch.name + "\\"
        level = level + 1
        self.__process_subbatches(parent_uid, sub_batch.subBatches, reference_unit_batch, path, level)


class S3OSIPIReader:
    @staticmethod
    def read_machine_status(full_path_to_file):
        data_location = 's3://{}'.format(full_path_to_file)
        df = pd.read_csv(data_location, parse_dates=['time'], infer_datetime_format=True)
        df.columns = ['Tag', 'Time', 'Status', 'Svalue', 'Pistatus', 'Flags']
        df.drop('Svalue', axis=1, inplace=True)
        df.drop('Pistatus', axis=1, inplace=True)
        df.drop('Flags', axis=1, inplace=True)
        return df


class FileOSIPIReader:
    @staticmethod
    def read_machine_status(full_path_to_file):
        df = pd.read_csv(full_path_to_file, parse_dates=['time'], infer_datetime_format=True)
        df.columns = ['Tag', 'Time', 'Status', 'Svalue', 'Pistatus', 'Flags']
        df.drop('Svalue', axis=1, inplace=True)
        df.drop('Pistatus', axis=1, inplace=True)
        df.drop('Flags', axis=1, inplace=True)
        return df


from .pg_utils import PGDataAccess, PGTSTableDataAccess
from .pi_utils import BoundaryType, PITagValues, PITagValue, OSIPIReader, PIDataUBSerializer, S3OSIPIReader, \
    FileOSIPIReader
from .s3_utils import Boto3ClientException, Boto3ClientFactory, S3SelectHelper
from .generic_utils import GenericUtils, DateTimeUtils



"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

import numpy as np
from aimltools.ts.preprocessing import DataPreparation
from collections import Counter, OrderedDict
import logging

logger = logging.getLogger(__name__)


class AnomalyDetectionValidation:
    
    def __init__(self):
        self.__TXT_HYPER_FLD = 'hyperparameters'
        self.__TXT_CLASSIFIER_FLD = 'classifier'

        self.__TXT_ACC = "accuracy (tp+tn/(tp+tn+fp+fn))"
        self.__TXT_TRUEN = "true negative rate (tn/(tn+fp))"
        self.__TXT_NEGATIVEPV = "negative pred. value (tn/(tn+fn))"
        self.__TXT_F1SCORE = "f1 score (2*(P*R)/(P+R))"
        self.__TXT_RECALL = "recall (tp/(tp+fn))"
        self.__TXT_PREC = "precision (tp/(tp+fp))"
        self.__TXT_FN = "anomalous detected as normal (false negatives)"
        self.__TXT_TP = "anomalous detected as anomalous (true positives)"
        self.__TXT_FP = "normal detected as anomalous (false positives)"
        self.__TXT_TN = "normal detected as normal (true negatives)"
        self.__TXT_ACC_WP = "accuracy - with perturbations"
        self.__TXT_ACC_NP = "accuracy - no perturbations"
        self.__TXT_CNT_WP = "count - with perturbations"
        self.__TXT_CNT_NP = "count - no perturbations"
        self.__TXT_SCORE_WP = "score - with perturbations"
        self.__TXT_SCORE_NP = "score - no perturbations"
        self.__TXT_PRED_WP = "pred - with perturbations"
        self.__TXT_PRED_NP = "pred - no perturbations"
        self.__TXT_MODEL = 'model'
    
    def get_fn(self):
        return self.__results[self.__TXT_FN]
    
    def get_tp(self):
        return self.__results[self.__TXT_TP]
    
    def get_fp(self):
        return self.__results[self.__TXT_FP]
    
    def get_tn(self):
        return self.__results[self.__TXT_TN]
    
    def get_precision(self):
        try:
            return self.__results[self.__TXT_TP] / (self.__results[self.__TXT_TP] + self.__results[self.__TXT_FP])
        except:
            logger.error("Cannot calculate {}: {} and {} are both zero".format(self.__TXT_PREC, self.__TXT_TP, self.__TXT_FP))
            return np.nan


    def get_recall(self):
        try:
            return self.__results[self.__TXT_TP] / (self.__results[self.__TXT_TP] + self.__results[self.__TXT_FN])
        except:
            logger.error("Cannot calculate {}: {} and {} are both zero".format(self.__TXT_RECALL, self.__TXT_TP, self.__TXT_FN))
            return np.nan


    def get_specificity(self):
        try:
            return self.__results[self.__TXT_TN] / (self.__results[self.__TXT_TN] + self.__results[self.__TXT_FP])
        except:
            logger.error("Cannot calculate {}: {} and {} are both zero".format(self.__TXT_TRUEN, self.__TXT_TN, self.__TXT_FP))
            return np.nan
    
    
    def get_negativepv(self):
        try:
            return self.__results[self.__TXT_TN] / (self.__results[self.__TXT_TN] + self.__results[self.__TXT_FN])
        except:
            logger.error("Cannot calculate {}: {} and {} are both zero".format(self.__TXT_NEGATIVEPV, self.__TXT_TN, self.__TXT_FN))
            return np.nan
    
    
    def get_f1score(self):
        p = self.get_precision()
        r = self.get_recall()
        try:
            return 2*(p*r)/(p+r)
        except:
            logger.error("Cannot calculate {}: {} and {} are both zero".format(self.__TXT_F1SCORE, self.__TXT_PREC, self.__TXT_RECALL))
            return np.nan
    
    
    def get_accuracy(self):
        t = self.__results[self.__TXT_TP] + self.__results[self.__TXT_TN]
        f = self.__results[self.__TXT_FP] + self.__results[self.__TXT_FN]
        try:
            return t/(t+f)
        except:
            logger.error("Cannot calculate {}: division by zero".format(self.__TXT_ACC))
            return np.nan


    @staticmethod
    def __count(preds, preds_perturb):
        cnt = dict(Counter(preds))
        cnt_perturbed = dict(Counter(preds_perturb))
        
        for i in [0, 1]:
            if not (i in cnt.keys()):
                cnt[i] = 0
            if not (i in cnt_perturbed.keys()):
                cnt_perturbed[i] = 0
        
        cnt = OrderedDict(sorted(cnt.items()))
        cnt_perturbed = OrderedDict(sorted(cnt_perturbed.items()))
        return [cnt, cnt_perturbed]
    
    
    def __create_result_dict(self, model_name, preds, preds_proba, preds_perturb, preds_proba_perturb):
        [cnt, cnt_perturbed] = AnomalyDetectionValidation.__count(preds, preds_perturb)
        
        self.__results = {}
        self.__results[self.__TXT_MODEL] = model_name
        self.__results[self.__TXT_PRED_NP] = preds
        self.__results[self.__TXT_PRED_WP] = preds_perturb
        self.__results[self.__TXT_SCORE_NP] = preds_proba
        self.__results[self.__TXT_SCORE_WP] = preds_proba_perturb
        self.__results[self.__TXT_CNT_NP] = cnt
        self.__results[self.__TXT_CNT_WP] = cnt_perturbed
        #self.__results[self.__TXT_ACC_NP] = cnt[0] / sum(cnt.values())
        #self.__results[self.__TXT_ACC_WP] = cnt_perturbed[1] / sum(cnt_perturbed.values())
        self.__results[self.__TXT_TN] = cnt[0]
        self.__results[self.__TXT_FP] = cnt[1]
        self.__results[self.__TXT_TP] = cnt_perturbed[1]
        self.__results[self.__TXT_FN] = cnt_perturbed[0]
        
        self.__results[self.__TXT_ACC] = self.get_accuracy()
        self.__results[self.__TXT_PREC] = self.get_precision()
        self.__results[self.__TXT_RECALL] = self.get_recall()
        self.__results[self.__TXT_TRUEN] = self.get_specificity()
        self.__results[self.__TXT_NEGATIVEPV] = self.get_negativepv()
        self.__results[self.__TXT_F1SCORE] = self.get_f1score()
        
    
    def get_results_summary(self):
        summary = self.__results.copy()
        summary.pop(self.__TXT_PRED_NP, None)
        summary.pop(self.__TXT_PRED_WP, None)
        summary.pop(self.__TXT_SCORE_NP, None)
        summary.pop(self.__TXT_SCORE_WP, None)
        return summary
        
    
    def validate(self, df_train, df_test, df_test_perturb, classifier_metadata_dict):
        hyperparams = dict()
        try:
            hyperparams = classifier_metadata_dict[self.__TXT_HYPER_FLD]
        except:
            pass

        try:
            df_train.drop(['cycle'], axis=1, inplace=True)
            df_test.drop(['cycle'], axis=1, inplace=True)
            df_test_perturb.drop(['cycle'], axis=1, inplace=True)
        except:
            pass

        dp = DataPreparation()
        df_train_filled = dp.fill_na(df_train)
        [scaler, df_train_norm] = dp.normalize(df_train_filled)

        lc = classifier_metadata_dict[self.__TXT_CLASSIFIER_FLD]
        lc.fit(X = df_train_norm, **hyperparams)
        
        df_test_filled = dp.fill_na(df_test)
        df_test_norm = scaler.transform(df_test_filled)
        preds = lc.predict(df_test_norm)
        preds_proba = lc.predict_proba(df_test_norm)

        df_test_perturb_filled = dp.fill_na(df_test_perturb)
        df_test_perturb_norm = scaler.transform(df_test_perturb_filled)
        preds_perturb = lc.predict(df_test_perturb_norm)
        preds_proba_perturb = lc.predict_proba(df_test_perturb_norm)

        self.__create_result_dict(str(type(lc).__name__), preds, preds_proba, preds_perturb, preds_proba_perturb)
        
        return self.__results
    

def validate(df_train_distance, df_test_distance, df_test_perturb_distance, classifier_metadata_dict):
    adv = AnomalyDetectionValidation()
    results = adv.validate(df_train_distance, df_test_distance, df_test_perturb_distance, 
                            classifier_metadata_dict = classifier_metadata_dict)
    return results



"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

"""
Sub-module of TSAnalysis

Contains classes and methods to denoise signals
"""

from tslearn.generators import random_walk_blobs, random_walks
import matplotlib.pyplot as plt
import pywt
from abc import ABCMeta, abstractmethod
import logging

logger = logging.getLogger(__name__)


# Defines denoiser mixins
class BaseMixin(metaclass = ABCMeta):
    """
    Defines pattern for child mixins
    """

    @abstractmethod
    def denoise_transform(self):
        """Mandatorily implement this method in all child mixins"""


class WaveletDenoiserMixin(BaseMixin):
    """
    Mixin class for denoising through discrete wavelet transform
    """
        
    def denoise_transform(self, signal, wavelet_type= 'sym4', level = None, 
                        threshold = 0.7, denoise_type = 'soft', axis = -1,
                        wavedec_mode = 'symmetric', threshold_mode = 'soft'):
        
        self.coeffs = pywt.wavedec(data = signal, wavelet = wavelet_type, 
                                    level = level, axis = axis, mode = wavedec_mode)
        
        for i in range(1, len(self.coeffs)):
            self.coeffs[i] = pywt.threshold(data = self.coeffs[i], 
                                            value = threshold * max(self.coeffs[i]),
                                            mode = threshold_mode)
            
        self.datarec = pywt.waverec(coeffs = self.coeffs, wavelet = wavelet_type)
        
        if signal.shape[0] != self.datarec.shape[0]:
            n = self.datarec.shape[0] - signal.shape[0]
            self.datarec = self.datarec[:-n]
            
        self.noise = signal - self.datarec
        return self.datarec, self.noise
    

# Defines base class
class DenoiserBaseClass(metaclass = ABCMeta):
    """
    Base class containing common functionality and abstract methods
    """
    
    def __init__(self, signal, denoise_type, **kwargs):
        self.signal = signal
        self.denoise_type = denoise_type
        self.kwargs = kwargs
    
    def __repr__(self):
        return """Denoise type = {}, \n
                Arguments = {}""".format(self.denoise_type, self.kwargs)
    
    @abstractmethod
    def denoise_method(self):
        """Mandatorily implement this method in all child classes"""


# Defines user interface
def denoise(signal, denoise_type = 'wavelet', return_noise = False, **kwargs):
    
    """
    Function that returns a denoised signal for an input signal

    if return_noise == True, also returns the noise component

    Parameters
    ----------
    signal : list or numpy.ndarray
            A list or numpy array of the input signal
    
    denoise_type: str
            Type of method for denoising
            Currently supported denoisers are:
            'wavelet' = Discrete wavelet transform
    
    return_noise: bool
            Flag to return the signal's noise component after denoising 
    
    kwargs: dict
            Keyword arguments to be passed depending on value of 
            'denoise_type'

    Returns
    -------
    signal_denoised: numpy.ndarray
            Denoised signal

    *if return_noise == True, also returns* 
    signal_noise: numpy.ndarray
            Noise extracted from signal during denoising
    """

    # Contains the denoise_type and its corresponding mixin
    # Add a new {denoise_type: mixin} key-value pair to 
    # register a new denoising method
    denoise_mixin_match_dict = {
                                'wavelet': WaveletDenoiserMixin
                                }

    supported = list(denoise_mixin_match_dict.keys())
    
    try:
        mixin_to_use = denoise_mixin_match_dict[denoise_type]
    except:
        raise ValueError("""
        Unsupported value '{}' for denoise_type. 
        Currently supported denoisers are - 
        {}""".format(denoise_type, supported))

    # Throws exception if mixin_to_use is not a child of BaseMixin
    assert issubclass(mixin_to_use, BaseMixin)

    # Defines child class from base class and required mixin
    class Denoiser(DenoiserBaseClass, mixin_to_use):
        
        def __init__(self, signal, denoise_type, **kwargs):
            super().__init__(signal, denoise_type, **kwargs)
        
        def denoise_method(self):
            signal_denoised, signal_noise = self.denoise_transform(signal, **self.kwargs)
            return signal_denoised, signal_noise

    # Instantiates child class
    instance = Denoiser(signal, denoise_type, **kwargs)
    signal_denoised, signal_noise = instance.denoise_method()

    if return_noise == True:
        logger.info("Output returned as (np.array(denoise_signal), np.array(noise))")
        return signal_denoised, signal_noise
    else:
        return signal_denoised
        


"""
Authors:    Francesco Gabbanini <gabbanini_francesco@lilly.com>, 
            Manjunath C Bagewadi <bagewadi_manjunath_c@lilly.com>, 
            Henson Tauro <tauro_henson@lilly.com> 
            (MQ IDS - Data Integration and Analytics)
License:    MIT
"""

import numpy as np
from scipy.stats import wasserstein_distance
import math

def find_position(p):
    if len(p)==1:
        return [0]
    else:
        q = np.sort(p)
        return [np.where(q == p[0])[0][0]]+find_position(p[1:])


def rolling_window(a, window, step_size):
    shape = a.shape[:-1] + (a.shape[-1] - window + 1 - step_size + 1, window)
    strides = a.strides + (a.strides[-1] * step_size,)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def compute_probs(data, n=10): 
    h, e = np.histogram(data, n)
    p = h/data.shape[0]
    return e, p


def kl_divergence(p, q): 
    return np.sum(p*np.log(p/q))


def js_divergence(p, q):
    m = (1./2.)*(p + q)
    return (1./2.)*kl_divergence(p, m) + (1./2.)*kl_divergence(q, m)


def compute_lehmer(w):
    
    fact = [math.factorial(i) for i in range(len(w)-1,0,-1)]+[0]
    q = find_position(w)
    return  np.sum(np.multiply(fact,q))


def compute_kl_divergence(train_sample, test_sample, n_bins=10): 
    """
    Computes the KL Divergence using the support intersection between two different samples
    """
    e, p = compute_probs(train_sample, n=n_bins)
    _, q = compute_probs(test_sample, n=e)

#     list_of_tuples = support_intersection(p, q)
#     p, q = get_probs(list_of_tuples)
    
    q = np.where(q==0,np.finfo(np.float32).eps,q)
    p = np.where(p==0,np.finfo(np.float32).eps,p)
    
    return kl_divergence(p, q)


def compute_js_divergence(train_sample, test_sample, n_bins=10): 
    """
    Computes the JS Divergence using the support intersection between two different samples
    """
    e, p = compute_probs(train_sample, n=n_bins)
    _, q = compute_probs(test_sample, n=e)
    
#     list_of_tuples = support_intersection(p, q)
#     p, q = get_probs(list_of_tuples)
    
    q = np.where(q==0,np.finfo(np.float32).eps,q)
    p = np.where(p==0,np.finfo(np.float32).eps,p)
    
    return js_divergence(p, q)


def PD_distance(p, q, metric = 'kullback', m = 3, step = 1):
     
    m_p = rolling_window(p,m, step)
    m_q = rolling_window(q,m, step)
    
    p_lv = np.array([compute_lehmer(xi) for xi in m_p])
    q_lv = np.array([compute_lehmer(xi) for xi in m_q])
    
    if metric == 'kullback':
        return compute_kl_divergence(p_lv,q_lv)
    
    if metric == 'wasserstein':
        return wasserstein_distance(p_lv,q_lv)


import logging
import pandas as pd
import ciso8601
import time
from datetime import datetime

logger = logging.getLogger(__name__)


class GenericUtils:
    @staticmethod
    def read_timestamp(FILENAME):             
        try:
            with open(FILENAME) as file:
                data = file.readlines()
            latest_timestamp = data[-1]
            file.close()
            return latest_timestamp
        except (Exception,IOError) as error:
            logger.info("Unable to read file: {}".format(error))
        finally:
            file.close()
    
    
    @staticmethod
    def write_timestamp(ts, FILENAME, mode = 'w'):
        try:
            with open(FILENAME, mode) as file:
                file.write(ts + '\n')
                file.close()
            return
        except (Exception,IOError) as error:
            logger.info("Unable to write file: {}".format(error))
        finally:
            file.close()
            
    @staticmethod
    def cls_to_df(df_train_ft, cls_scores_list, cls_reslts_list, eqm_id, model_id, output_cols):
        #eqpm_id     integer NOT NULL,
        #modl_id     varchar(50) NOT NULL,
        #evnt_time   timestamp(6) with time zone NOT NULL,
        #score       real NOT NULL,
        #is_anomaly boolean NOT NULL,
        #feat_count 
        
        result_list = []
        i = 0
        for index, row in df_train_ft.iterrows():
            is_anomalous = False
            if cls_reslts_list[i] == 1:
                is_anomalous = True
            
            lst = df_train_ft.columns[row.notnull()]
            feature_count = len(lst)
            #rc_features = "|".join(lst)
            tmp = [eqm_id, model_id, index, cls_scores_list[i], is_anomalous, feature_count] 
            result_list.append(tmp)
            i = i + 1
    
        df_result = pd.DataFrame(result_list, columns = output_cols)
        return df_result


    @staticmethod
    def cls_res_to_df(df_train_ft, cls_res_list, model_id, score_thr, output_cols):
        result_list = []
        i = 0
        for index, row in df_train_ft.iterrows():
            res_str = "NORMAL"
            if cls_res_list[i] > score_thr:
                res_str = "ANOMALOUS"
            
            lst = df_train_ft.columns[row.notnull()]
            feature_count = len(lst)
            rc_features = "|".join(lst)
            tmp = [model_id, index, cls_res_list[i], res_str, rc_features, score_thr, feature_count] 
            result_list.append(tmp)
            i = i + 1
    
        df_result = pd.DataFrame(result_list, columns = output_cols)
        return df_result


class DateTimeUtils:
    @staticmethod
    def string_to_datetime_ciso8601(time_str):
        return ciso8601.parse_datetime(time_str)


    @staticmethod
    def datetime_to_epoch(dt):
        return dt.timestamp()


    @staticmethod
    def string_to_epoch_ciso8601(time_str):
        return DateTimeUtils.datetime_to_epoch(DateTimeUtils.string_to_datetime_ciso8601(time_str))


    @staticmethod
    def epoch_to_string(epoch_dt, time_str_format):
        return time.strftime(time_str_format, time.localtime(epoch_dt))


    @staticmethod
    def epoch_to_string_tz(epoch_dt, time_str_format, tzone):
        return datetime.fromtimestamp(epoch_dt, tzone).strftime(time_str_format)


    @staticmethod
    def datetime_to_string(dt, time_str_format):
        return dt.strftime(time_str_format)




from setuptools import setup, find_packages
from codecs import open
import numpy
import os

# thanks Pipy for handling markdown now
ROOT = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(ROOT, 'README.md'), encoding="utf-8") as f:
    README = f.read()

import aimltools
VERSION = aimltools.__version__

# The following block
# is copied from sklearn as a way to avoid cythonizing source
# files in source dist
# We need to import setuptools before because it monkey-patches distutils
#from distutils.command.sdist import sdist
#cmdclass = {'sdist': sdist}

setup(
    name='dia-aiml',  # How you named your package folder (MyLib)
    description='A Python toolkit for the analysis of machine data',  # Short description about your library
    long_description=README,
    long_description_content_type='text/markdown',
    license='MIT',
    include_dirs=[numpy.get_include()],
    packages=find_packages(),
    install_requires=[
        'boto3',
        'botocore',
        'ciso8601',
        'combo',
        'deprecation',
        'Keras',
        'numpy',
        'pandas',
        'psycopg2',
        'PyNomaly',
        'pyod',
        #'pythonnet',
        'PyWavelets',
        'PyYAML',
        'requests',
        'scikit-learn',
        'scipy',
        'shap',
        'tabulate',
        #'tensorflow',
        'tqdm',
        'tslearn',
    ],
    version=VERSION,
    url="https://github.com/EliLillyCo/MQIT_DIA_AIML_Toolkit",
    author='Francesco Gabbanini, Manjunath Bagewadi, Henson Tauro',  # Type in your name
    author_email='fgabbanini@yahoo.it',  # Type in your E-Mail
    #cmdclass=cmdclass
)


import psycopg2
from psycopg2 import extras
import logging
import pandas as pd
from deprecation import deprecated

logger = logging.getLogger(__name__)

class PGDataAccess:
    
    def __init__(self,USER,PASS,HOST,PORT,DB,SSL,SSL_CERT):
        self.__USER = USER
        self.__PASS = PASS
        self.__HOST = HOST
        self.__PORT = PORT
        self.__DB = DB
        self.__SSL = SSL
        self.__SSL_CERT = SSL_CERT

    def _get_connection(self):
        return self.__connection

    def connect(self):
        try:
            self.__connection = psycopg2.connect(user=self.__USER, password=self.__PASS, host=self.__HOST, port=self.__PORT, database=self.__DB,sslmode=self.__SSL, sslrootcert=self.__SSL_CERT)
        except (Exception, psycopg2.Error) as error:
            logger.error("Error while connecting to PostgreSQL {}".format(error))
    
    
    def disconnect(self):
        self.__connection.close()


    def execute_simple_delete(self, table_name, key_fld, key_val):
        sql_statement = "DELETE FROM {} WHERE {} = %s".format(table_name, key_fld)
        try:
            cur = self.__connection.cursor()
            cur.execute(sql_statement, [key_val])
            self.__connection.commit()
        except Exception as error:
            logger.error("Error while deleting data : {}".format(error), exc_info=True)
        finally:
            cur.close()


    def execute_select(self, sql_statement, params):
        try:
            data = pd.read_sql(sql_statement, self.__connection, params = params)
            return data
        except Exception as error:
            logger.error("Error while reading data : {}".format(error), exc_info=True)

    def execute_simple_select(self, table_name, key_fld, key_val):
        sql_statement = "SELECT * FROM {} WHERE {} = %(key_val)s".format(table_name, key_fld)
        try:
            data = pd.read_sql(sql_statement, self.__connection, params={'key_val': key_val})
            return data
        except Exception as error:
            logger.error("Error while reading data : {}".format(error), exc_info=True)

    @deprecated(deprecated_in="0.9.0", removed_in="1.0.0",
                details="Use PGDataAccess.write_results_bulk(...) from pg_utils.py instead")
    def write_results(self, df, table_name):
        str_col_list = ", ".join(df.columns)
        str_col_placeholders = ", ".join(['%s'] * len(df.columns))
        stmt = "INSERT INTO {} ({}) VALUES({})".format(table_name, str_col_list, str_col_placeholders)
        record_count = 0
        try:
            cur = self.__connection.cursor()
            for index, row in df.iterrows():
                cur.execute(stmt, row.tolist())
                record_count = record_count + cur.rowcount
        except Exception as error:
            logger.error("Error while writing data : {}".format(error), exc_info=True)
            record_count = 0
        finally:
            self.__connection.commit()
            cur.close()
            return record_count


    def write_results_bulk(self, df, table_name):
        str_col_list = ", ".join(df.columns)
        stmt = "INSERT INTO {} ({}) VALUES %s".format(table_name, str_col_list)
        record_count = len(df)
        write_list = list(df.itertuples(index=False, name=None))

        try:
            cur = self.__connection.cursor()
            extras.execute_values(cur, stmt, write_list)
        except Exception as error:
            logger.error("Error while writing data : {}".format(error), exc_info=True)
            record_count = 0
        finally:
            self.__connection.commit()
            cur.close()
            return record_count


    def vacuum_table(self, table_name, vacuum_type = 'partial'):

        if vacuum_type == 'partial':
            vacuum_type = ''

        sql_statement = "VACUUM {} {}".format(vacuum_type, table_name)
    
        try:
            self._get_connection().set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
            cur = self._get_connection().cursor()
            cur.execute(sql_statement)
            self._get_connection().commit()
            logger.info("Successfully executed VACUUM on table {}".format(table_name))
            self._get_connection().set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_DEFAULT)
        except Exception as error:
            logger.error("Error while performing VACUUM on table {} : {}".format(table_name, error), exc_info=True)
        finally:
            cur.close()

    
    def reindex_table(self):
        table_name = self.__table_name
        sql_statement = "REINDEX TABLE {}".format(table_name)
        try:
            cur = self._get_connection().cursor()
            cur.execute(sql_statement)
            self._get_connection().commit()
            logger.info("Successfully executed REINDEX TABLE on table {}".format(table_name))
        except Exception as error:
            logger.error("Error while performing REINDEX TABLE on table {} : {}".format(table_name, error), exc_info=True)
        finally:
            cur.close()


"""
Used to manage PostgreSQL tables collecting time series values, having following structure:
            VNAME, VTIME, VVALUE, VSTATUS
            - VNAME: name of variable being measured in given record
            - VTIME: time when variable is measured
            - VVALUE: value of VNAME at VTIME (if numeric)
            - VSTATUS: value of VNAME at VTIME (if string or category)
"""
class PGTSTableDataAccess(PGDataAccess):
    """
        Parameters:
        - table_name: name of table
        - name_fld: name of field reporting variable name
        - time_fld: name of field reporting timestamp
        - value_fld: name of field reporting value (if numeric)
        - status_fld: name of field reporting value (if string)
    """
    def setup_mappings(self, table_name, name_fld, time_fld, value_fld, status_fld):
        self.__table_name = table_name
        self.__name_fld = name_fld
        self.__time_fld = time_fld
        self.__value_fld = value_fld
        self.__status_fld = status_fld

    """
    Parameters:
        - columns: name of table columns to be read
        - start: time filter (select records where self.__time_fld > start)
        - end: time filter (select records where self.__time_fld <= end)
        - vnames_list: filter on variable names (select records where self.__name_fld is in vnames_list)
        - order_by: list of list to specify order of records (e.g., [['event_time', 'asc']])
        - status_fld: name of field reporting value (if string) 
    """
    def read_data(self, start, end, vnames_list=[], columns=[], order_by=[]):
        try:
            cols = '*'
            if len(columns) > 0:
                cols = ','.join(columns)

            if len(vnames_list) > 0:
                vn = "('{}')".format("','".join(vnames_list))
                sql = "select {} from {} where (({} > %(start_date)s and {} <= %(end_date)s) and {} in {})".format(
                    cols,
                    self.__table_name,
                    self.__time_fld,
                    self.__time_fld,
                    self.__name_fld,
                    vn
                    )
            else:
                sql = "select {} from {} where (({} > %(start_date)s and {} <= %(end_date)s))".format(
                    cols,
                    self.__table_name,
                    self.__time_fld,
                    self.__time_fld)
            
            if len(order_by) > 0:
                oblist = []
                for itm in order_by:
                    oblist.extend(["{} {}".format(itm[0], itm[1])])
                sql += " order by " + (", ".join(oblist))

            param_dict = {'start_date': start, 'end_date': end}
            data = pd.read_sql(sql, self._get_connection(), params=param_dict)
            return data
        except Exception as error:
            logger.error("Error while reading data : {}".format(error))


    def read_data_count(self, start, end):
        try:
            sql = "select count(*) from {} where {} between %(start_date)s and %(end_date)s".format(self.__table_name, self.__time_fld)
            param_dict = {'start_date': start, 'end_date': end}
            data = pd.read_sql(sql, self._get_connection(), params=param_dict)
            return data.iloc[0][0]
        except Exception as error:
            logger.error("Error while reading data count: {}".format(error))


    def read_cycle_count(self, start, end):
        try:
            sql = "select count(distinct {}) from {} where {} between %(start_date)s and %(end_date)s".format(self.__time_fld, self.__table_name, self.__time_fld)
            param_dict = {'start_date': start, 'end_date': end}
            data = pd.read_sql(sql, self._get_connection(), params=param_dict)
            return data.iloc[0][0]
        except Exception as error:
            logger.error("Error while reading cycle count: {}".format(error))


    def delete_data(self, start, end):
        sql_statement = "DELETE FROM {} WHERE {} > %(start_date)s and {} <= %(end_date)s".format(self.__table_name, self.__time_fld, self.__time_fld)
        try:
            cur = self._get_connection().cursor()
            cur.execute(sql_statement, {"start_date": start, "end_date": end})
            self._get_connection().commit()
        except Exception as error:
            logger.error("Error while deleting data : {}".format(error), exc_info=True)
        finally:
            cur.close()
